from anvil import ANVIL, CONFIG
from anvil.lib.lib import *
from anvil.lib.schemas import AddonObject
from PIL import Image

__all__ = [
    "UIBindingType",
    "UIElementType",
    "UIAnchor",
    "UITextAlignment",
    "UIAnimType",
    "UIEasing",
    "UIElementTrigger",
    "UIFontSize",
    "UI",
]

class UIBindingType(StrEnum):
    View = "view"
    Global = "global"
    Collection = "collection"
    CollectionDetails = "collection_details"
    NONE = "none"


class UIElementType(StrEnum):
    Image = "image"
    Panel = "panel"
    Label = "label"
    Screen = "screen"
    Factory = "factory"
    StackPanel = "stack_panel"
    Grid = "grid"
    Custom = "custom"
    Button = "button"


class UIAnchor(StrEnum):
    Center = "center"
    TopLeft = "top_left"
    TopMiddle = "top_middle"
    TopRight = "top_right"
    LeftMiddle = "left_middle"
    RightMiddle = "right_middle"
    BottomLeft = "bottom_left"
    BottomMiddle = "bottom_middle"
    BottomRight = "bottom_right"


class UITextAlignment(StrEnum):
    Center = "center"
    Left = "left"
    Right = "right"


class UIAnimType(StrEnum):
    Alpha = "alpha"
    Clip = "clip"
    Color = "color"
    Flip_book = "flip_book"
    Offset = "offset"
    Size = "size"
    UV = "uv"
    Wait = "wait"
    Aseprite_flip_book = "aseprite_flip_book"


class UIEasing(StrEnum):
    Linear = "linear"
    Spring = "spring"
    InQuad = "in_quad"
    OutQuad = "out_quad"
    InOutQuad = "in_out_quad"
    InCubic = "in_cubic"
    OutCubic = "out_cubic"
    InOutCubic = "in_out_cubic"
    InQuart = "in_quart"
    OutQuart = "out_quart"
    InOutQuart = "in_out_quart"
    InQuint = "in_quint"
    OutQuint = "out_quint"
    InOutQuint = "in_out_quint"
    InSine = "in_sine"
    OutSine = "out_sine"
    InOutSine = "in_out_sine"
    InExpo = "in_expo"
    OutExpo = "out_expo"
    InOutExpo = "in_out_expo"
    InCirc = "in_circ"
    OutCirc = "out_circ"
    InOutCirc = "in_out_circ"
    InBounce = "in_bounce"
    OutBounce = "out_bounce"
    InOutBounce = "in_out_bounce"
    InBack = "in_back"
    OutBack = "out_back"
    InOutBack = "in_out_back"
    InElastic = "in_elastic"
    OutElastic = "out_elastic"
    InOutElastic = "in_out_elastic"


class UIElementTrigger(StrEnum):
    Title = "title"
    Actionbar = "actionbar"
    NONE = "None"


class UIFontSize(StrEnum):
    Normal = "normal"
    Small = "small"
    Large = "large"
    ExtraLarge = "extra_large"


# Subclasses ==========================================================


class _UIBinding:
    def __init__(self) -> None:
        self._content = {}

    def binding_name(self, binding_name: str):
        self._content["binding_name"] = binding_name
        return self

    def binding_name_override(self, binding_name_override: str):
        self._content["binding_name_override"] = binding_name_override
        return self

    def binding_type(self, binding_type: UIBindingType = UIBindingType.View):
        self._content["binding_type"] = binding_type.value
        return self

    def binding_collection_name(self, binding_collection_name: str):
        self._content["binding_collection_name"] = binding_collection_name
        return self

    def binding_collection_prefix(self, binding_collection_prefix: str):
        self._content["binding_collection_prefix"] = binding_collection_prefix
        return self

    def source_control_name(self, source_control_name: str):
        self._content["source_control_name"] = source_control_name
        return self

    def source_property_name(self, source_property_name: str):
        self._content["source_property_name"] = source_property_name
        return self

    def target_property_name(self, target_property_name: str):
        self._content["target_property_name"] = target_property_name
        return self

    @property
    def resolve_sibling_scope(
        self,
    ):
        self._content["resolve_sibling_scope"] = True
        return self

    def binding_condition(self, binding_condition: str):
        self._content["binding_condition"] = binding_condition
        return self


class _UIButtonMapping:
    def __init__(self) -> None:
        self._content = {}

    def from_button_id(self, from_button_id: str):
        self._content["from_button_id"] = from_button_id
        return self

    def to_button_id(self, to_button_id: str):
        self._content["to_button_id"] = to_button_id
        return self

    def mapping_type(self, mapping_type: str):
        self._content["mapping_type"] = mapping_type
        return self

    def ignored(self, ignored: bool):
        self._content["ignored"] = ignored
        return self

    @property
    def button_up_right_of_first_refusal(self):
        self._content["button_up_right_of_first_refusal"] = True
        return self


class _UIModifications:
    def __init__(self) -> None:
        self._content = {}

    def remove(self, control_name: str):
        self._content["control_name"] = control_name
        self._content["operation"] = "remove"
        return self

    def insert_front(self, control_name: str):
        element = _UIElement(control_name)
        self._content["array_name"] = "controls"
        self._content["operation"] = "insert_front"
        self._content["value"] = element
        return element


class _UIElement:
    def __init__(self, element_name: str) -> None:
        self._element_name = element_name
        self._textures = []
        self.element = {}
        self._buttons: list[_UIButtonMapping] = []
        self._bindings: list[_UIBinding] = []
        self._controls: list[_UIElement] = []
        self._modifications: list[_UIModifications] = []

    @property
    def should_steal_mouse(self):
        self.element["should_steal_mouse"] = True
        return self

    @property
    def absorbs_input(self):
        self.element["absorbs_input"] = True
        return self

    def visible(self, visible: bool | str):
        self.element["visible"] = visible
        return self

    def fill(self, fill: bool | str):
        self.element["fill"] = fill
        return self

    def ignored(self, ignored: bool | str):
        self.element["ignored"] = ignored
        return self

    def enabled(self, enabled: bool | str):
        self.element["enabled"] = enabled
        return self

    def text(self, text: str):
        self.element["text"] = text
        return self

    def localize(self, localize: bool):
        self.element["localize"] = localize
        return self

    def renderer(self, renderer: str):
        self.element["renderer"] = renderer
        return self

    def type(self, type: UIElementType):
        self.element["type"] = type.value
        return self

    def orientation(self, orientation: str):
        self.element["orientation"] = orientation
        return self

    def anchor(self, anchor_from: UIAnchor, anchor_to: UIAnchor):
        self.element["anchor_from"] = anchor_from.value if isinstance(anchor_from, UIAnchor) else anchor_from
        self.element["anchor_to"] = anchor_to.value if isinstance(anchor_to, UIAnchor) else anchor_to
        return self

    def text_alignment(self, text_alignment: UITextAlignment):
        self.element["text_alignment"] = text_alignment.value
        return self

    def color(self, color: tuple[float, float, float]):
        self.element["color"] = color
        return self

    def font_size(self, font_size: UIFontSize):
        self.element["font_size"] = font_size.value
        return self

    def layer(self, layer: int):
        self.element["layer"] = layer
        return self

    def locked_control(self, locked_control: str):
        self.element["locked_control"] = locked_control
        return self

    def font_scale_factor(self, font_scale_factor: int):
        self.element["font_scale_factor"] = font_scale_factor
        return self

    def texture(self, texture: str, scale_factor: int = 1, *nineslice_size: int):
        if not "$" in texture:
            if FileExists(os.path.join("assets", "textures", "ui", f"{texture}.png")):
                self.element["texture"] = os.path.join("textures", "ui", texture)
                self._textures.append(texture)
            else:
                ANVIL.config.Logger.file_exist_error(f"{texture}.png", os.path.join("assets", "textures", "ui"))
        else:
            self.element["texture"] = texture

        if len(nineslice_size) > 0 or scale_factor != 1:
            File(
                f"{texture}.json",
                {
                    "nineslice_size": nineslice_size[0] * scale_factor if len(nineslice_size) == 1 else [i * scale_factor for i in nineslice_size],
                    "base_size": [i * scale_factor for i in Image.open(os.path.join("assets", "textures", "ui", f"{texture}.png")).size],
                },
                os.path.join(CONFIG.RP_PATH, "textures", "ui"),
                "w",
            )
        return self

    def texture_key(self, key: str, texture: str, scale_factor: int = 1, *nineslice_size: int):
        if FileExists(os.path.join("assets", "textures", "ui", f"{texture}.png")):
            self.element[f"${key}"] = os.path.join("textures", "ui", texture)
            self._textures.append(texture)
            if len(nineslice_size) > 0 or scale_factor != 1:
                File(
                    f"{texture}.json",
                    {
                        "nineslice_size": nineslice_size,
                        "base_size": [i * scale_factor for i in Image.open(os.path.join("assets", "textures", "ui", f"{texture}.png")).size],
                    },
                    os.path.join(CONFIG.RP_PATH, "textures", "ui"),
                    "w",
                )
        else:
            ANVIL.config.Logger.file_exist_error(f"{texture}.png", os.path.join("assets", "textures", "ui"))
        return self

    def aseprite_texture(self, texture: str):
        if FileExists(os.path.join("assets", "textures", "ui", f"{texture}.png")):
            self.texture(texture)
            CopyFiles(
                os.path.join("assets", "textures", "ui"),
                os.path.join(CONFIG.RP_PATH, "textures", "ui"),
                f"{texture}.json",
            )
        else:
            ANVIL.config.Logger.file_exist_error(f"{texture}.png", os.path.join("assets", "textures", "ui"))
        return self

    def aseprite_key(self, key: str, texture: str):
        if FileExists(os.path.join("assets", "textures", "ui", f"{texture}.png")):
            self.element[f"${key}"] = os.path.join("textures", "ui", texture)
            self._textures.append(texture)
            CopyFiles(
                os.path.join("assets", "textures", "ui"),
                os.path.join(CONFIG.RP_PATH, "textures", "ui"),
                f"{texture}.json",
            )
        else:
            ANVIL.config.Logger.file_exist_error(f"{texture}.png", os.path.join("assets", "textures", "ui"))
        return self

    def keys(self, key, value):
        self.element[f"${key}"] = value.value if isinstance(value, StrEnum) else value
        return self

    def property_bag(self, **properties):
        if "property_bag" not in self.element:
            self.element["property_bag"] = {}
        self.element["property_bag"].update({f"#{k}": v for k, v in properties.items()})
        return self

    @property
    def binding(self):
        if "bindings" not in self.element:
            self.element["bindings"] = []
        bind = _UIBinding()
        self._bindings.append(bind)
        return bind

    @property
    def button_mapping(self):
        if "button_mappings" not in self.element:
            self.element["button_mappings"] = []
        bind = _UIButtonMapping()
        self._buttons.append(bind)
        return bind

    @property
    def modification(self):
        if "modifications" not in self.element:
            self.element["modifications"] = []
        mod = _UIModifications()
        self._modifications.append(mod)
        return mod

    def control_ids(self, id: str, element: str):
        if "control_ids" not in self.element:
            self.element["control_ids"] = {}
        self.element["control_ids"][id] = element

        return self

    def controls(self, element_name: str):
        if "controls" not in self.element:
            self.element["controls"] = []

        ctrl = _UIElement(element_name)
        self._controls.append(ctrl)
        return ctrl

    def factory(self, name, id, element):
        if not id == "control_name":
            self.element["factory"] = {"name": name, "control_ids": {id: element}}
        else:
            self.element["factory"] = {"name": name, id: element}

        return self

    def size(self, size: str | tuple):
        self.element["size"] = size
        return self

    def tiled(self, bool: bool | str):
        self.element["tiled"] = bool
        return self

    def max_size(self, max_size: str | tuple):
        self.element["max_size"] = max_size
        return self

    def min_size(self, min_size: str | tuple):
        self.element["min_size"] = min_size
        return self

    def clip_ratio(self, clip_ratio: float | str):
        self.element["clip_ratio"] = clip_ratio
        return self

    def clip_direction(self, clip_direction: float | str):
        self.element["clip_direction"] = clip_direction
        return self

    def clip_pixelperfect(self, clip_pixelperfect: bool):
        self.element["clip_pixelperfect"] = clip_pixelperfect
        return self

    def maximum_grid_items(self, maximum_grid_items: int):
        self.element["maximum_grid_items"] = maximum_grid_items
        return self

    def line_padding(self, line_padding: int):
        self.element["line_padding"] = line_padding
        return self

    def uv(self, uv: tuple | str):
        self.element["uv"] = uv
        return self

    def uv_size(self, uv_size: tuple):
        self.element["uv_size"] = uv_size
        return self

    def alpha(self, alpha: str | tuple):
        self.element["alpha"] = alpha
        return self

    def offset(self, offset: str | tuple):
        self.element["offset"] = offset
        return self

    def clip_offset(self, clip_offset: tuple):
        self.element["clip_offset"] = clip_offset
        return self

    def keep_ratio(self, keep_ratio: bool):
        self.element["keep_ratio"] = keep_ratio
        return self

    def allow_clipping(self, allow_clipping: bool):
        self.element["allow_clipping"] = allow_clipping
        return self

    def grid_item_template(self, grid_item_template: bool):
        self.element["grid_item_template"] = grid_item_template
        return self

    def grid_dimension_binding(self, grid_dimension_binding: bool):
        self.element["grid_dimension_binding"] = grid_dimension_binding
        return self

    def grid_dimensions(self, grid_dimensions: tuple):
        self.element["grid_dimensions"] = grid_dimensions
        return self

    def grid_position(self, grid_position: tuple):
        self.element["grid_position"] = grid_position
        return self

    def grid_rescaling_type(self, grid_rescaling_type: str):
        self.element["grid_rescaling_type"] = grid_rescaling_type
        return self

    def grid_fill_direction(self, grid_fill_direction: str):
        self.element["grid_fill_direction"] = grid_fill_direction
        return self

    def collection_index(self, collection_index: int):
        self.element["collection_index"] = collection_index
        return self

    def collection_name(self, collection_name: bool):
        self.element["collection_name"] = collection_name
        return self

    def shadow(self, shadow: bool = False):
        self.element["shadow"] = shadow
        return self

    def use_anchored_offset(self, use_anchored_offset: bool = False):
        self.element["use_anchored_offset"] = use_anchored_offset
        return self

    def clips_children(self, clips_children: bool):
        self.element["clips_children"] = clips_children
        return self

    @property
    def enable_scissor_test(self):
        self.element["enable_scissor_test"] = True
        return self

    @property
    def focus_enabled(self):
        self.element["focus_enabled"] = True
        return self

    @property
    def propagate_alpha(self):
        self.element["propagate_alpha"] = True
        return self

    def variables(self, requires: str, **kwargs):
        if "variables" not in self.element:
            self.element["variables"] = []
        vars = {"requires": requires}
        for k, v in kwargs.items():
            vars.update({f"${k}": v})
        self.element["variables"].append(vars)
        return self

    @property
    def queue(self):
        for texture in self._textures:
            CopyFiles(
                os.path.join("assets", "textures", "ui"),
                os.path.join(
                    CONFIG.RP_PATH,
                    "textures",
                    "ui",
                ),
                f"{texture}.png",
            )
        for bind in self._bindings:
            self.element["bindings"].append(bind._content)
        for button in self._buttons:
            self.element["button_mappings"].append(button._content)
        for mod in self._modifications:
            if "value" in mod._content:
                mod._content["value"] = mod._content["value"].queue

            self.element["modifications"].append(mod._content)
        for ctrl in self._controls:
            self.element["controls"].append(ctrl.queue)
        return {self._element_name: self.element}


class _UIAnimationElement:
    def __init__(self, animation_name: str) -> None:
        self._animation_name = animation_name
        self.animation = {}

    def anim_type(self, anim_type: UIAnimType):
        self.animation["anim_type"] = anim_type.value
        return self

    def duration(self, duration: int | str):
        self.animation["duration"] = duration
        return self

    def next(self, next: str):
        self.animation["next"] = next
        return self

    def destroy_at_end(self, destroy_at_end: str):
        self.animation["destroy_at_end"] = destroy_at_end
        return self

    def play_event(self, play_event: str):
        self.animation["play_event"] = play_event
        return self

    def end_event(self, end_event: str):
        self.animation["end_event"] = end_event
        return self

    def start_event(self, start_event: str):
        self.animation["start_event"] = start_event
        return self

    def reset_event(self, reset_event: str):
        self.animation["reset_event"] = reset_event
        return self

    def easing(self, easing: UIEasing):
        self.animation["easing"] = easing.value
        return self

    def from_(self, from_: str | tuple):
        self.animation["from"] = from_
        return self

    def to(self, to: str | tuple):
        self.animation["to"] = to
        return self

    def initial_uv(self, initial_uv: tuple = (0, 0)):
        self.animation["initial_uv"] = initial_uv
        return self

    def fps(self, fps: int):
        self.animation["fps"] = fps
        return self

    def frame_count(self, frame_count: int):
        self.animation["frame_count"] = frame_count
        return self

    def frame_step(self, frame_step: int):
        self.animation["frame_step"] = frame_step
        return self

    @property
    def reversible(self):
        self.animation["reversible"] = True
        return self

    @property
    def resettable(self):
        self.animation["resettable"] = True
        return self

    @property
    def scale_from_starting_alpha(self):
        self.animation["scale_from_starting_alpha"] = True
        return self

    @property
    def activated(self):
        self.animation["activated"] = True
        return self

    @property
    def queue(self):
        return {self._animation_name: self.animation}

    def __str__(self) -> str:
        return self._animation_name


class _UICreditsConstructor:
    def __init__(self, hud_screen: "_UIScreen", credits_panel: _UIElement, credits_duration: int = 30) -> None:
        self._hud_screen = hud_screen

        anim = self._hud_screen._anvil_animation.add_animation("credits_scroll")
        anim.anim_type(UIAnimType.Offset)
        anim.duration(credits_duration)
        anim.from_((0, "160%")).to((0, "-160%"))
        anim.destroy_at_end("hud_title_text")

        self._credits_panel = credits_panel.controls("credits_text@anvil_common.stack_panel")
        self._credits_panel.offset("@anvil_animations.credits_scroll")

        self._element_count = 0

    def add_image(self, texture_name: str, scale_factor: int = 1, anchor = UIAnchor.Center, *nineslice_size: int):
        """Adds an image to the credits panel.

        Args:
            texture_name (str): The name of the texture.
            scale_factor (int, optional): The scale factor of the texture. Defaults to 1.
            anchor (UIAnchor, optional): The anchor of the image. Defaults to UIAnchor.Center.

        Returns:
            _UICreditsConstructor: The credits constructor.
        """
        image = self._credits_panel.controls(f"panel_{self._element_count}@anvil_common.panel").size(("100%", "100%c")).controls(
            f"image_{self._element_count}@anvil_common.image"
        )
        image.texture(texture_name, scale_factor, *nineslice_size)
        image.size((50, 50))
        if anchor != UIAnchor.Center:
            image.anchor(anchor, anchor)

        self._element_count += 1
        return self

    def add_title(
        self,
        text: str,
        font_size: UIFontSize = UIFontSize.Normal,
        color: tuple[float] = (1, 1, 1),
        text_alignment: UITextAlignment = UITextAlignment.Center,
    ):
        """Adds a title to the credits panel.

        Args:
            text (str): The text of the title.
            font_size (UIFontSize, optional): The font size of the title. Defaults to UIFontSize.Normal.
            color (tuple[float], optional): The color of the title. Defaults to (1, 1, 1).
            text_alignment (UITextAlignment, optional): The text alignment of the title. Defaults to UITextAlignment.Center.

        Returns:
            _UICreditsConstructor: The credits constructor.
        """
        title = self._credits_panel.controls(f"panel_{self._element_count}@anvil_common.panel").size(("100%", "100%c")).controls(
            f"title_{self._element_count}@anvil_common.label"
        )
        title.text(text)
        if font_size != UIFontSize.Normal:
            title.font_size(font_size)
        if color != (1, 1, 1):
            title.color(color)
        if text_alignment != UITextAlignment.Center:
            title.text_alignment(text_alignment)
            title.size(("100%", "default"))
            
        self._element_count += 1
        return self

    def add_space(self, size: int | str):
        """Adds a space to the credits panel.

        Args:
            size (int | str): The size of the space.

        Returns:
            _UICreditsConstructor: The credits constructor.
        """
        space = self._credits_panel.controls(f"space_{self._element_count}@anvil_common.panel")
        space.size((0, size))
        self._element_count += 1
        return self

    def add_section(
        self,
        section_title: str,
        section_names: list[str],
        title_font_size: UIFontSize = UIFontSize.Normal,
        title_color: tuple[float] = (1, 1, 1),
        names_font_size: UIFontSize = UIFontSize.Normal,
        names_color: tuple[float] = (1, 1, 1),
    ):
        # Section Setup
        section = self._credits_panel.controls(f"credits_section_{self._element_count}@anvil_common.stack_panel")
        section.orientation("horizontal")
        section.size(("100%", "100%c"))

        # Section title
        title = section.controls(f"panel_title_{self._element_count}@anvil_common.panel").size(("45%", "100%c")).controls(f"title_{self._element_count}@anvil_common.label")
        title.text(section_title)
        title.size(("100%", "default"))
        title.text_alignment(UITextAlignment.Right)
        if title_font_size != UIFontSize.Normal:
            title.font_size(title_font_size)
        if title_color != (1, 1, 1):
            title.color(title_color)

        # Section separator
        section.controls(f"section_sep_{self._element_count}@anvil_common.panel").size(("10%", 0))

        ## Section names
        names = section.controls(f"names_{self._element_count}@anvil_common.stack_panel")
        names.size(("45%", "100%c"))
        for i, n in enumerate(section_names):
            name = names.controls(f"name_{i}_{self._element_count}@anvil_common.label")
            name.text(n)
            name.text_alignment(UITextAlignment.Left)
            if names_font_size != UIFontSize.Normal:
                name.font_size(names_font_size)
            if names_color != (1, 1, 1):
                name.color(names_color)

        self._element_count += 1

# UI files ============================================================
class _UIVariables(AddonObject):
    _extension = ".json"
    _path = os.path.join(CONFIG.RP_PATH, "ui")

    def __init__(self) -> None:
        super().__init__("_global_variables")

    def add_variable(self, variable, value):
        self._content.update({variable: value})


class _UIDefs(AddonObject):
    _extension = ".json"
    _path = os.path.join(CONFIG.RP_PATH)
    
    def __init__(self) -> None:
        super().__init__("_ui_defs")
        self.do_not_shorten
        self._files = []

    def add_file(self, path: str):
        self._files.append(path)

    @property
    def queue(self):
        self.content({"ui_defs": self._files})
        return super().queue("ui")


class _UIAnimation(AddonObject):
    _extension = ".json"
    _path = os.path.join(CONFIG.RP_PATH, "ui")

    def __init__(self, name: str, namespace: str, defs: _UIDefs) -> None:
        super().__init__(name)
        self.name = name
        self._defs = defs
        self.namespace = namespace
        self._animations: list[_UIAnimationElement] = []
        self._content = {
            "namespace": namespace,
        }

    def add_animation(self, animation_name: str):
        self._animations.append(_UIAnimationElement(animation_name))
        return self._animations[-1]

    def queue(self, directory: str = ""):
        for anim in self._animations:
            self._content.update(anim.queue)

        self._defs.add_file(os.path.join("ui", directory, f"{self.name}{_UIScreen._extension}"))
        return super().queue(directory)


class _UIScreen(AddonObject):
    _extension = ".json"
    _path = os.path.join(CONFIG.RP_PATH, "ui")

    def __init__(self, name: str, namespace: str, anvil_animation: _UIAnimation, variables: _UIVariables, defs: _UIDefs) -> None:
        super().__init__(name)
        self._content = {
            "namespace": namespace,
        }
        self.name = name
        self.namespace = namespace
        self._elements: list[_UIElement] = []
        self._anvil_animation = anvil_animation
        self._variables = variables
        self._defs = defs
        self._ignored_title_texts = []
        self._ignored_actionbar_texts = []
        self._hides_hud = []
        self.do_not_shorten

    def add_element(
        self,
        element_name: str,
        trigger: UIElementTrigger = UIElementTrigger.NONE,
        keyword: str = None,
        hides_hud: bool = False,
    ):
        # Element name
        new_element = _UIElement(element_name)

        match trigger:
            case UIElementTrigger.Title:
                if not keyword is None:
                    self._variables.add_variable(f"$anvil.{element_name}.text", f"{CONFIG.NAMESPACE}:{keyword}")
                    self._variables.add_variable(f"$anvil.{element_name}.bool", f"(not ((#title_text - $anvil.{element_name}.text) = #title_text))")
                    self._ignored_title_texts.append(f"$anvil.{element_name}.text")
                    if hides_hud:
                        self._hides_hud.append(f"$anvil.{element_name}.text")

                    new_element = _UIElement(f"{element_name}@anvil_common.title_binding")
                    new_element.keys("binding_text", f"$anvil.{element_name}.bool")

                factory = self.add_element(f"{element_name}_factory")
                factory.type(UIElementType.Panel)
                factory.factory("hud_title_text_factory", "hud_title_text", f"{element_name}_element@{self.namespace}.{element_name}")
                self._elements.append(factory)

            case UIElementTrigger.Actionbar:
                if not keyword is None:
                    self._variables.add_variable(f"$anvil.{element_name}.text", f"{CONFIG.NAMESPACE}:{keyword}")
                    self._variables.add_variable(f"$anvil.{element_name}.bool", f"(not(($text - $anvil.{element_name}.text) = $text))")
                    self._ignored_actionbar_texts.append(f"$anvil.{element_name}.text")

                    new_element = _UIElement(f"{element_name}@anvil_common.actionbar_binding")
                    new_element.keys("binding_text", f"$anvil.{element_name}.bool")

                factory: _UIElement = self.add_element(f"{element_name}_factory")
                factory.type(UIElementType.Panel)
                factory.factory(
                    "hud_actionbar_text_factory",
                    "hud_actionbar_text",
                    f"{element_name}_element@anvil_hud.{element_name}",
                )

        self._elements.append(new_element)
        return new_element

    def queue(self, directory: str = ""):
        for element in self._elements:
            self._content.update(element.queue)

        if self._content != {"namespace": self.namespace}:
            self._defs.add_file(os.path.join("ui", directory, f"{self.name}{_UIScreen._extension}"))
            return super().queue(directory)

# UI Screens ==========================================================
# Vanilla
class _HUDScreen(_UIScreen):
    def __init__(self, anvil_animation: _UIAnimation, variables: _UIVariables, defs: _UIDefs) -> None:
        super().__init__("hud_screen", "hud", anvil_animation, variables, defs)
        # Disable HUD
        self.root_panel = self.add_element("root_panel")
        self.anvil_element = self.root_panel.modification.insert_front("anvil@anvil_hud.anvil_hud")

        self.hud_title_text = self.add_element("hud_title_text")
        self.hud_title_text.binding.binding_name("#hud_title_text_string").binding_name_override("#text")
        self.hud_title_text.binding.binding_name("#hud_subtitle_text_string").binding_name_override("#subtext")

        self.hud_actionbar_text = self.add_element("hud_actionbar_text")
        self.hud_actionbar_text.keys("text", "$actionbar_text")

        variables.add_variable("$anvil.show.text", f"{CONFIG.NAMESPACE}:show")
        variables.add_variable("$anvil.hide.text", f"{CONFIG.NAMESPACE}:hide")
        self._ignored_title_texts = ["$anvil.show.text"]
        self._ignored_actionbar_texts = ["$anvil.hide.text"]
        self._hides_hud = ["$anvil.hide.text"]

    def disable_mouse(self):
        self.add_element("hud_screen@common.base_screen").should_steal_mouse.absorbs_input
        self.root_panel.modification.remove("curor_rend")
        return self

    def disable_mob_effect(self):
        self.root_panel.modification.remove("mob_effects_renderer")
        return self

    def disable_helpers(self):
        self.root_panel.modification.remove("left_helpers")
        self.root_panel.modification.remove("right_helpers")
        self.root_panel.modification.remove("emote_expediate_helpers")
        return self

    def disable_hotbar(self):
        self.root_panel.modification.remove("centered_gui_elements")
        self.root_panel.modification.remove("centered_gui_elements_at_bottom_middle")
        self.root_panel.modification.remove("centered_ridingvr_gui_elements")
        self.root_panel.modification.remove("not_centered_gui_elements")
        self.root_panel.modification.remove("exp_rend")
        self.root_panel.modification.remove("exp_rend_resizable")
        return self

    def disable_sidebar(self):
        self.root_panel.modification.remove("sidebar")
        return self

    def queue(self, directory: str = ""):
        return super().queue()


class _NPCScreen(_UIScreen):
    def __init__(self, anvil_animation: _UIAnimation, variables: _UIVariables, defs: _UIDefs) -> None:
        super().__init__("npc_interact_screen", "npc_interact", anvil_animation, variables, defs)
        self._npc_screen = self.add_element("npc_screen@common.base_screen").keys("screen_content", "anvil_npc.npc_screen_chooser")

    def add_element(self, element_name: str, trigger: UIElementTrigger = UIElementTrigger.NONE, keyword: str = None):
        return super().add_element(element_name, trigger, keyword, False)
    
    def queue(self, directory: str = ""):
        return super().queue()


# Anvil
class _AnvilHUDScreen(_UIScreen):
    def __init__(self, anvil_animation: _UIAnimation, variables: _UIVariables, defs: _UIDefs) -> None:
        super().__init__("hud", "anvil_hud", anvil_animation, variables, defs)
        self.anvil_hud: _UIElement = super().add_element("anvil_hud")
        self.anvil_hud.type(UIElementType.Panel)

    def add_element(
        self,
        element_name: str,
        trigger: UIElementTrigger = UIElementTrigger.NONE,
        keyword: str = None,
        hides_hud: bool = False,
    ):
        if trigger is not UIElementTrigger.NONE:
            self.anvil_hud.controls(f"{element_name}_instance@anvil_hud.{element_name}_factory")

        return super().add_element(element_name, trigger, keyword, hides_hud)

    def add_image_panel(
        self,
        name: str,
        background: str = "transparent",
        direction_out: bool = True,
        zoom_in: float = 1,
        zoom_wait: float = 6,
        zoom_out: float = 1,
    ):
        # element
        image_element = self.add_element(name, UIElementTrigger.Title, name)
        image_element.type(UIElementType.Panel)
        image_element.size(("100%", "100%"))
        image_element.layer(100)
        image_element.anchor(UIAnchor.Center, UIAnchor.Center)

        if not background == "transparent":
            background_image = image_element.controls(f"{name}_background")
            background_image.type(UIElementType.Image)
            background_image.texture(background)
            background_image.size(("100%", "100%"))
            background_image.keep_ratio(False)

        image = image_element.controls(f"{name}_image")
        image.type(UIElementType.Image)
        image.texture(name)
        image.anchor(UIAnchor.Center, UIAnchor.Center)
        image.offset((0, "-20px"))

        tip = image_element.controls(f"{name}_tip")
        tip.type(UIElementType.Label)
        tip.text_alignment(UITextAlignment.Center)
        tip.size(("80%", "default"))
        tip.offset((0, "100px"))
        tip.text("#text").shadow(True).font_size(UIFontSize.Large)
        tip.binding.binding_name("#hud_subtitle_text_string").binding_name_override("#text")

        zoom_steps = []
        if direction_out:
            zoom_steps = ["30%", "40%", "50%"]
        else:
            zoom_steps = ["400%", "70%", "60%"]

        # animations
        self._variables.add_variable(f"$anvil.image.{name}_in", zoom_in)
        self._variables.add_variable(f"$anvil.image.{name}_wait", zoom_wait)
        self._variables.add_variable(f"$anvil.image.{name}_out", zoom_out)

        image_zoom_in = self._anvil_animation.add_animation(f"{name}_zoom_in")
        image_zoom_in.anim_type(UIAnimType.Size)
        image_zoom_in.easing(UIEasing.Linear)
        image_zoom_in.from_((zoom_steps[0], zoom_steps[0])).to((zoom_steps[1], zoom_steps[1]))
        image_zoom_in.duration(f"$anvil.image.{name}_in")
        image_zoom_in.next(f"@anvil_animations.{name}_zoom_wait")

        image_zoom_wait = self._anvil_animation.add_animation(f"{name}_zoom_wait")
        image_zoom_wait.anim_type(UIAnimType.Wait)
        image_zoom_wait.duration(f"$anvil.image.{name}_wait")
        image_zoom_wait.next(f"@anvil_animations.{name}_zoom_out")
        image_zoom_out = self._anvil_animation.add_animation(f"{name}_zoom_out")

        image_zoom_out.anim_type(UIAnimType.Size)
        image_zoom_out.easing(UIEasing.Linear)
        image_zoom_out.from_((zoom_steps[1], zoom_steps[1])).to((zoom_steps[2], zoom_steps[2]))
        image_zoom_out.duration(f"$anvil.image.{name}_out")

        image_alpha_in = self._anvil_animation.add_animation(f"{name}_alpha_in")
        image_alpha_in.anim_type(UIAnimType.Alpha)
        image_alpha_in.easing(UIEasing.OutQuad)
        image_alpha_in.from_(0)
        image_alpha_in.to(1)
        image_alpha_in.duration(f"$anvil.image.{name}_in")
        image_alpha_in.next(f"@anvil_animations.{name}_alpha_wait")
        image_alpha_wait = self._anvil_animation.add_animation(f"{name}_alpha_wait")
        image_alpha_wait.anim_type(UIAnimType.Wait)
        image_alpha_wait.duration(f"$anvil.image.{name}_wait")
        image_alpha_wait.next(f"@anvil_animations.{name}_alpha_out")
        image_alpha_out = self._anvil_animation.add_animation(f"{name}_alpha_out")
        image_alpha_out.anim_type(UIAnimType.Alpha)
        image_alpha_out.easing(UIEasing.InOutSine)
        image_alpha_out.from_(1)
        image_alpha_out.to(0)
        image_alpha_out.duration(f"$anvil.image.{name}_out")
        image_alpha_out.destroy_at_end("hud_title_text")

        image.size(f"@anvil_animations.{name}_zoom_in" if zoom_in > 0 else ("50%", "50%"))
        image.alpha(f"@anvil_animations.{name}_alpha_in" if zoom_in > 0 else 1)

        return image_element

    # Layer 100
    def add_logo(self):
        self.add_image_panel("logo")

    # Layer 101
    def add_blinking_screen(self):
        # animation
        blink_fade_in = self._anvil_animation.add_animation("blink_fade_in")
        blink_fade_in.anim_type(UIAnimType.Alpha)
        blink_fade_in.easing(UIEasing.InOutSine)
        blink_fade_in.from_(0).to(1)
        self._variables.add_variable("$anvil.blink.fade_in", 1)
        blink_fade_in.duration("$anvil.blink.fade_in")
        blink_fade_in.next("@anvil_animations.blink_fade_stay")

        blink_fade_stay = self._anvil_animation.add_animation("blink_fade_stay")
        blink_fade_stay.anim_type(UIAnimType.Wait)
        self._variables.add_variable("$anvil.blink.fade_stay", 1)
        blink_fade_stay.duration("$anvil.blink.fade_stay")
        blink_fade_stay.next("@anvil_animations.blink_fade_out")

        blink_fade_out = self._anvil_animation.add_animation("blink_fade_out")
        blink_fade_out.anim_type(UIAnimType.Alpha)
        blink_fade_out.easing(UIEasing.InOutSine)
        blink_fade_out.from_(1).to(0)
        self._variables.add_variable("$anvil.blink.fade_out", 1)
        blink_fade_out.duration("$anvil.blink.fade_out")
        # blink_fade_out.destroy_at_end('hud_title_text')
        # element
        blink_element = self.add_element("blink_element", UIElementTrigger.Title, "blink")
        blink_element.type(UIElementType.Image)
        blink_element.layer(101)
        blink_element.texture("black_element")
        blink_element.anchor(UIAnchor.Center, UIAnchor.Center)
        blink_element.size(("300%", "300%"))
        blink_element.alpha("@anvil_animations.blink_fade_in")

    # Layer 101
    def add_black_bars(self):
        black_bars_in = self._anvil_animation.add_animation("black_bars_in")
        black_bars_in.anim_type(UIAnimType.Size)
        black_bars_in.easing(UIEasing.InOutSine)
        black_bars_in.from_(("300%", "300%")).to(("300%", "100%"))
        self._variables.add_variable("$anvil.black_bars.in", 1.6)
        black_bars_in.duration("$anvil.black_bars.in")
        black_bars_in.next("@anvil_animations.black_bars_wait")

        black_bars_wait = self._anvil_animation.add_animation("black_bars_wait")
        black_bars_wait.anim_type(UIAnimType.Wait)
        self._variables.add_variable("$anvil.black_bars.stay", 1)
        black_bars_wait.duration("$anvil.black_bars.stay")

        black_bars_out = self._anvil_animation.add_animation("black_bars_out")
        black_bars_out.anim_type(UIAnimType.Size)
        black_bars_out.easing(UIEasing.InOutSine)
        black_bars_out.from_(("300%", "100%")).to(("300%", "300%"))
        self._variables.add_variable("$anvil.black_bars.out", 1.6)
        black_bars_out.duration("$anvil.black_bars.out")
        # black_bars_out.destroy_at_end('hud_title_text')

        # element
        black_bars = self.add_element("black_bars_element", UIElementTrigger.Actionbar)
        black_bars.type(UIElementType.Image)
        black_bars.texture("black_bars")
        black_bars.keep_ratio(False)
        black_bars.anchor(UIAnchor.Center, UIAnchor.Center)
        black_bars.layer(101)
        black_bars.keys("anim_size", ("300%", "300%"))
        black_bars.keys("text", "$actionbar_text")
        black_bars.variables(
            "$anvil.black_bars_in.bool",
            anim_size="@anvil_animations.black_bars_in",
        )
        black_bars.variables(
            "$anvil.black_bars_out.bool",
            anim_size="@anvil_animations.black_bars_out",
        )
        black_bars.size("$anim_size")

        self._variables.add_variable(f"$anvil.black_bars_in.text", f"{CONFIG.NAMESPACE}:bars_in")
        self._variables.add_variable(f"$anvil.black_bars_in.bool", f"($text = $anvil.black_bars_in.text)")
        self._variables.add_variable(f"$anvil.black_bars_out.text", f"{CONFIG.NAMESPACE}:bars_out")
        self._variables.add_variable(f"$anvil.black_bars_out.bool", f"($text = $anvil.black_bars_out.text)")

        self._ignored_actionbar_texts.extend(
            [
                "($text = $anvil.black_bars_in.text)",
                "($text = $anvil.black_bars_out.text)",
            ]
        )


    def credits_screen_constructor(self, credits_duration: int = 30):
        credits = self.add_element("credits", UIElementTrigger.Title, "credits")
        credits.type(UIElementType.Panel)
        credits.layer(100)

        background_image = credits.controls("credits_background@anvil_common.image")
        background_image.texture("black_element")
        background_image.size(("100%", "100%"))
        background_image.keep_ratio(False)
        background_image.layer(-1)

        return _UICreditsConstructor(self, credits, credits_duration)


    def queue(self, directory: str = ""):
        return super().queue("anvil")


class _AnvilNPCScreen(_UIScreen):
    def __init__(self, anvil_animation: _UIAnimation, variables: _UIVariables, defs: _UIDefs) -> None:
        super().__init__("npc", "anvil_npc", anvil_animation, variables, defs)
        self._ignored_panel_texts = []

        self._npc_screen_chooser = self.add_element("npc_screen_chooser").type(UIElementType.Panel)
        vanilla = self._npc_screen_chooser.controls("vanilla@npc_interact.npc_screen_contents").layer(500)
        vanilla.binding.binding_type(UIBindingType.Global).binding_name("#title_text").binding_name_override("#title_text")
        vanilla.binding.binding_type(UIBindingType.Global).binding_name("#dialogtext").binding_name_override("#dialogtext")
        vanilla.binding.binding_type(UIBindingType.View).source_property_name("$anvil.npc_screen.vanilla").target_property_name("#visible")

    def add_element(self, element_name: str, trigger: UIElementTrigger = UIElementTrigger.NONE, keyword: str = None):
        return super().add_element(element_name, trigger, keyword, False)
    
    def queue(self, directory: str = ""):
        self._variables.add_variable("$anvil.npc_screen.vanilla", f'({" and ".join(self._ignored_panel_texts)})')
        return super().queue("anvil")


class _AnvilCommon(_UIScreen):
    def __init__(self, anvil_animation: _UIAnimation, variables: _UIVariables, defs: _UIDefs) -> None:
        super().__init__("common", "anvil_common", anvil_animation, variables, defs)
        self.basic_components()
        self.image_label()
        self.title_actionbar()
        self.scoreboard_retrieve()
        self.complex_components()

    def add_element(self, element_name: str, trigger: UIElementTrigger = UIElementTrigger.NONE, keyword: str = None):
        return super().add_element(element_name, trigger, keyword, False)
    
    def basic_components(self):
        # ---------------------------
        # Label
        # Must supply your own text
        # ---------------------------
        label = self.add_element("label")
        label.type(UIElementType.Label)
        label.text_alignment(UITextAlignment.Center)

        # ---------------------------
        # Image
        # Must supply your own texture
        # ---------------------------
        image = self.add_element("image")
        image.type(UIElementType.Image)

        # ---------------------------
        # Stack Panel
        # ---------------------------
        stack_panel = self.add_element("stack_panel")
        stack_panel.type(UIElementType.StackPanel)

        # ---------------------------
        # Stack Panel
        # ---------------------------
        panel = self.add_element("panel")
        panel.type(UIElementType.Panel)

    def complex_components(self):
        # npc model renderer (for npc screen)
        npc_renderer = self.add_element("npc_renderer")
        npc_renderer.type(UIElementType.Panel)
        npc_renderer.size(("100%", "100%"))
        npc_renderer.clips_children(True)
        npc_model = npc_renderer.controls("npc_model")
        npc_model.type(UIElementType.Custom)
        npc_model.layer(5)
        npc_model.renderer("actor_portrait_renderer")
        npc_model.size(("120%", "120%"))
        npc_model.enable_scissor_test
        npc_model.binding.binding_type(UIBindingType.Collection).binding_collection_name("skins_collection").binding_name("#skin_index")

        # player model renderer
        player_renderer = self.add_element("player_renderer")
        player_renderer.type(UIElementType.Panel)
        player_renderer.size(("80%", "80%"))
        player_renderer.clips_children(True)
        player_model = player_renderer.controls("player_model")
        player_model.type(UIElementType.Custom)
        player_model.renderer("live_player_renderer")
        player_model.size(("80%", "80%"))
        player_model.enable_scissor_test
        player_model.offset((0, "50%y"))

    def image_label(self):
        # ---------------------------
        # Image Label
        # Must supply your own text key
        # Must supply your own texture
        # ---------------------------
        image_label = self.add_element("image_label@anvil_common.image")
        image_label.size(("100%c + 8px", "100%c + 4px"))
        image_label.keys("text", "PLACEHOLDER TEXT")
        image_label.keys("localize", False)
        label = image_label.controls("label@anvil_common.label")
        label.text("$text")
        label.localize("$localize")

        # ---------------------------
        # Image Label Binding
        # Must supply your own control name key that has a text binding
        # Must supply your own texture
        # ---------------------------
        image_label_binding = self.add_element("image_label_binding@anvil_common.image")
        image_label_binding.size(("100%c + 8px", "100%c + 4px"))
        label_binding = image_label_binding.controls("text@anvil_common.label")
        label_binding.text("#text")
        label_binding.layer(1)
        label_binding.binding.binding_type(UIBindingType.View).source_control_name("$control_name").source_property_name("#text").target_property_name("#text")

    def title_actionbar(self):
        title_binding = self.add_element("title_binding")
        title_binding.property_bag(title_text="", subtitle_text="")
        title_binding.binding.binding_name("#hud_title_text_string").binding_name_override("#title_text")
        title_binding.binding.binding_name("#hud_subtitle_text_string").binding_name_override("#subtitle_text")
        title_binding.binding.binding_type(UIBindingType.View).source_property_name("$binding_text").target_property_name("#visible")

        actionbar_binding = self.add_element("actionbar_binding")
        actionbar_binding.keys("text", "$actionbar_text")
        actionbar_binding.visible("$binding_text")

    def scoreboard_retrieve(self):
        """Use ``retrieve_score`` as the element. A few variables must be passed.
        To retrieve the score value, you must call the source_control_name using the element based on``retrieve_score``, then call the property ``score`` for the int value, or ``text`` for the string value.

        - index: this is the index of the score in the sidebar scoreboard.
        - score_offset: the number to subtract from the score.
        - color [optional]: The RGB tuple color of the text.
        """
        scoreboard_score = self.add_element("scoreboard_score_element@anvil_common.label")
        scoreboard_score.text("#text")
        scoreboard_score.color("$color")
        scoreboard_score.shadow("$shadow")
        scoreboard_score.layer(1)
        scoreboard_score.binding.binding_name("#player_score_sidebar").binding_type(UIBindingType.Collection).binding_collection_name("scoreboard_scores")
        scoreboard_score.binding.binding_type(UIBindingType.View).source_property_name("('§z' + ((#player_score_sidebar * 1) - $score_offset))").target_property_name("#text")
        scoreboard_score.binding.binding_type(UIBindingType.View).source_property_name("((#player_score_sidebar * 1) - $score_offset)").target_property_name("#score")

        retrieve_score = self.add_element("retrieve_score@anvil_common.stack_panel")
        retrieve_score.size(("100%c", "100%c"))
        retrieve_score.keys("index", 0)
        retrieve_score.keys("shadow", False)
        retrieve_score.keys("score_offset", 0)
        retrieve_score.keys("color", (1, 1, 1))
        retrieve_score.keys("name", "('score_text_' + $index)")
        retrieve_score.collection_name("scoreboard_scores")
        retrieve_score.controls("$name@scoreboard_score_element").collection_index("$index")
        retrieve_score.visible(False)
        retrieve_score.binding.binding_type(UIBindingType.View).source_control_name("$name").source_property_name("#text").target_property_name("#text")
        retrieve_score.binding.binding_type(UIBindingType.View).source_control_name("$name").source_property_name("#score").target_property_name("#score")

    def queue(self, directory: str = ""):
        return super().queue("anvil")


class _AnvilAnimations(_UIAnimation):
    def __init__(self, defs: _UIDefs) -> None:
        super().__init__("animations", "anvil_animations", defs)

    @property
    def queue(self):
        return super().queue("anvil")


class UI:
    _extension = {0: ".json", 1: ".json"}

    def __init__(self) -> None:
        self._defs = _UIDefs()
        self.variables = _UIVariables()

        self.animations_screen = _AnvilAnimations(self._defs)

        self.hud_screen = _HUDScreen(self.animations_screen, self.variables, self._defs)
        self.anvil_hud_screen = _AnvilHUDScreen(self.animations_screen, self.variables, self._defs)

        self.npc_screen = _NPCScreen(self.animations_screen, self.variables, self._defs)
        self.anvil_npc_screen = _AnvilNPCScreen(self.animations_screen, self.variables, self._defs)

        self.anvil_common = _AnvilCommon(self.animations_screen, self.variables, self._defs)

        self._screens: list[_UIScreen] = [
            self.hud_screen,
            self.anvil_hud_screen,
            self.npc_screen,
            self.anvil_npc_screen,
            self.anvil_common
        ]

    def add_ui_screen(self, filename, namespace):
        self._screens.append(_UIScreen(filename, namespace, self.animations_screen, self.variables, self._defs))
        return self._screens[-1]

    def queue(self):
        self.anvil_hud_screen._variables.add_variable("$anvil.empty.text", "")
        self.anvil_hud_screen._ignored_title_texts.append("$anvil.empty.text")
        self.anvil_hud_screen._ignored_actionbar_texts.append("$anvil.empty.text")
        # Manage Titles, Subtitles and actionbars
        ignored_title_texts = self.hud_screen._ignored_title_texts
        ignored_title_texts.extend(self.anvil_hud_screen._ignored_title_texts)

        ignored_actionbar_texts = self.hud_screen._ignored_actionbar_texts
        ignored_actionbar_texts.extend(self.anvil_hud_screen._ignored_actionbar_texts)
        
        hides_hud = self.hud_screen._hides_hud

        # Add to ANVIL queue
        for screen in self._screens:
            ignored_actionbar_texts.extend(screen._ignored_actionbar_texts)
            ignored_title_texts.extend(screen._ignored_title_texts)
            hides_hud.extend(screen._hides_hud)

        self.anvil_hud_screen._variables.add_variable(
            f"$anvil.title_visible.bool",
            f"({'(' * len(ignored_title_texts)}#text - " + " - ".join(f"{var})" for var in ignored_title_texts) + " = #text)",
        )
        self.anvil_hud_screen._variables.add_variable(
            f"$anvil.actionbar_visible.bool",
            f"({'(' * len(ignored_actionbar_texts)}$text - " + " - ".join(f"{var})" for var in ignored_actionbar_texts) + " = $text)",
        )

        self.hud_screen.hud_title_text.binding.binding_type(UIBindingType.View).source_property_name(
            f"$anvil.title_visible.bool"
        ).target_property_name("#visible")
        self.hud_screen.hud_actionbar_text.visible(f"$anvil.actionbar_visible.bool")

        source_prop = f"(#hud_visible and ({'(' * len(hides_hud)}#text - " + " - ".join(f"{var})" for var in hides_hud) + " = #text))"

        self.hud_screen.root_panel.binding.binding_name("#hud_visible")
        self.hud_screen.root_panel.binding.binding_name("#hud_title_text_string").binding_name_override("#text")
        self.hud_screen.root_panel.binding.binding_type(UIBindingType.View).source_property_name(source_prop).target_property_name("#visible")

        for screen in self._screens:
            screen.queue("anvil")

        self.animations_screen.queue
        self.variables.queue()

        self._defs.queue



import json
from enum import StrEnum

from anvil.api.types import Identifier, coordinate, coordinates
from anvil.lib.lib import clamp, normalize_180


class Target(StrEnum):
    """
    Enumeration for the types of targets that can be selected in Minecraft.
    """

    P = "@p"
    R = "@r"
    A = "@a"
    E = "@e"
    S = "@s"
    C = "@c"
    V = "@v"
    Initiator = "@initiator"


class Style(StrEnum):
    """
    Enumeration for the different styles of text in the game.
    """

    Black: str = "§0"
    DarkBlue: str = "§1"
    DarkGreen: str = "§2"
    DarkAqua: str = "§3"
    DarkRed: str = "§4"
    DarkPurple: str = "§5"
    Gold: str = "§6"
    Gray: str = "§7"
    DarkGray: str = "§8"
    Blue: str = "§9"
    Green: str = "§a"
    Aqua: str = "§b"
    Red: str = "§c"
    Purple: str = "§d"
    Yellow: str = "§e"
    White: str = "§f"
    MineconGold: str = "§g"
    Obfuscated: str = "§k"
    Bold: str = "§l"
    Italic: str = "§o"
    Reset: str = "§r"


class FogCameraLocation(StrEnum):
    """
    Enumeration for the different locations of the fog camera in the game.
    """

    Air = "air"
    Lava = "lava"
    Lava_resistance = "lava_resistance"
    Powder_snow = "powder_snow"
    Water = "water"
    Weather = "weather"


class RenderDistanceType(StrEnum):
    """
    Enumeration for the different types of render distances in the game.
    """

    Render = "render"
    Fixed = "fixed"


class LootPoolType(StrEnum):
    """
    Enumeration for the different types of loot pools.
    """

    Empty = "empty"
    Item = "item"
    LootTable = "loot_table"


class Slots(StrEnum):
    """
    Enumeration for the different types of inventory slots in the game.
    """

    Mainhand = "slot.weapon.mainhand"
    Offhand = "slot.weapon.offhand"
    Head = "slot.armor.head"
    Chest = "slot.armor.chest"
    Legs = "slot.armor.legs"
    Feet = "slot.armor.feet"
    Hotbar = "slot.hotbar"
    Inventory = "slot.inventory"
    Enderchest = "slot.enderchest"
    Saddle = "slot.saddle"
    Armor = "slot.armor"
    EquippedChest = "slot.chest"
    Equippable = "slot.equippable"
    Container = "slots.container"
    Body = "slot.armor.body"


class Gamemodes(StrEnum):
    """
    Enumeration for the different types of game modes.
    """

    Adventure = "adventure"
    Creative = "creative"
    Default = "default"
    Survival = "survival"
    Spectator = "spectator"

    A = Adventure
    C = Creative
    D = Default
    S = Survival


class ScoreboardOperator(StrEnum):
    """
    Enumeration for the different types of mathematical operators.
    """

    Less = "<"
    LessEqual = "<="
    Equals = "="
    Greater = ">"
    GreaterEqual = ">="


class Dimension(StrEnum):
    """
    Enumeration for the different dimensions in the game.
    """

    Overworld = "overworld"
    Nether = "nether"
    End = "the_end"


class FillMode(StrEnum):
    Replace = "replace"
    Outline = "outline"
    Hollow = "hollow"
    Destroy = "destroy"
    Keep = "keep"


class MusicRepeatMode(StrEnum):
    Once = "play_once"
    Loop = "loop"


class Difficulty(StrEnum):
    """
    Enumeration for the different levels of game difficulty.
    """

    Peaceful = "peaceful"
    Easy = "easy"
    Normal = "normal"
    Hard = "hard"
    Hardcore = "hardcore"


class DamageCause(StrEnum):
    """
    Enumeration for the different causes of damage in the game.
    """

    All = "all"
    Anvil = "anvil"
    EntityAttack = "entity_attack"
    BlockExplosion = "block_explosion"
    Contact = "contact"
    Drowning = "drowning"
    EntityExplosion = "entity_explosion"
    Fall = "fall"
    FallingBlock = "falling_block"
    Fatal = "fatal"
    Fire = "fire"
    Firetick = "firetick"
    FlyIntoWall = "fly_into_wall"
    Lava = "lava"
    Magic = "magic"
    Nothing = "none"
    Override = "override"
    Piston = "piston"
    Projectile = "projectile"
    SonicBoom = "sonic_boom"
    Stalactite = "stalactite"
    Stalagmite = "stalagmite"
    Starve = "starve"
    Suffocation = "suffocation"
    SelfDestruct = "self_destruct"
    Thorns = "thorns"
    Void = "void"
    Wither = "wither"


class Effects(StrEnum):
    """
    Enumeration for the different types of effects in the game.
    """

    Speed = "speed"
    Slowness = "slowness"
    Haste = "haste"
    MiningFatigue = "mining_fatigue"
    Strength = "strength"
    InstantHealth = "instant_health"
    InstantDamage = "instant_damage"
    JumpBoost = "jump"
    Hunger = "hunger"
    Saturation = "saturation"
    Regeneration = "regeneration"
    Weakness = "weakness"
    Levitation = "levitation"
    Wither = "wither"
    Poison = "poison"
    Absorption = "absorption"
    Invisibility = "invisibility"
    SlowFalling = "slow_falling"
    Nausea = "nausea"
    NightVision = "night_vision"
    WaterBreathing = "water_breathing"
    ConduitPower = "conduit_power"
    Resistance = "resistance"
    # Added in 1.21
    Infested = "infested"
    Oozing = "oozing"
    Weaving = "weaving"
    WindCharged = "wind_charged"
    RaidOmen = "raid_omen"
    TrailOmen = "trail_omen"


class ScoreboardOperation(StrEnum):
    """
    Enumeration for the different operations that can be performed on a scoreboard.
    """

    Addition = "+="
    Subtraction = "-="
    Multiplication = "*="
    Division = "/="
    Modulus = "%="
    Assign = "="
    Min = "<"
    Max = ">"
    Swaps = "><"


class Anchor(StrEnum):
    """
    Enumeration representing the two anchor points that can be used in Minecraft: the feet and the eyes.
    """

    Feet = "feet"
    Eyes = "eyes"


class CameraShakeType(StrEnum):
    """
    Enumeration representing the types of camera shakes that can occur in Minecraft.
    """

    positional = "positional"
    rotational = "rotational"


class MaskMode(StrEnum):
    """
    Enumeration representing the different mask modes that can be applied in Minecraft.
    """

    replace = "replace"
    masked = "masked"


class CloneMode(StrEnum):
    """
    Enumeration representing the different modes that can be used when cloning in Minecraft.
    """

    force = "force"
    move = "move"
    normal = "normal"


class InputPermissions(StrEnum):
    """
    Enumeration representing the different input permissions that can be set for a player in Minecraft.
    """

    Camera = "camera"
    Movement = "movement"


class CameraPresets(StrEnum):
    FirstPerson = "minecraft:first_person"
    ThirdPerson = "minecraft:third_person"
    ThirdPersonFront = "minecraft:third_person_front"
    Free = "minecraft:free"
    FollowOrbit = "minecraft:follow_orbit"


class CameraEasing(StrEnum):
    Linear = "linear"
    Spring = "spring"
    InQuad = "in_quad"
    OutQuad = "out_quad"
    InOutQuad = "in_out_quad"
    InCubic = "in_cubic"
    OutCubic = "out_cubic"
    InOutCubic = "in_out_cubic"
    InQuart = "in_quart"
    OutQuart = "out_quart"
    InOutQuart = "in_out_quart"
    InQuint = "in_quint"
    OutQuint = "out_quint"
    InOutQuint = "in_out_quint"
    InSine = "in_sine"
    OutSine = "out_sine"
    InOutSine = "in_out_sine"
    InExpo = "in_expo"
    OutExpo = "out_expo"
    InOutExpo = "in_out_expo"
    InCirc = "in_circ"
    OutCirc = "out_circ"
    InOutCirc = "in_out_circ"
    InBounce = "in_bounce"
    OutBounce = "out_bounce"
    InOutBounce = "in_out_bounce"
    InBack = "in_back"
    OutBack = "out_back"
    InOutBack = "in_out_back"
    InElastic = "in_elastic"
    OutElastic = "out_elastic"
    InOutElastic = "in_out_elastic"


class TimeSpec(StrEnum):
    DAY = "day"
    SUNRISE = "sunrise"
    NOON = "noon"
    SUNSET = "sunset"
    NIGHT = "night"
    MIDNIGHT = "midnight"


class Selector:
    """
    A class used to construct a target selector for Minecraft commands. The class offers various methods to set target
    parameters such as its type, name, count, coordinates, distance, volume, scores, rotation, permissions, and gamemode.
    """

    def __init__(self, target: Target = Target.S) -> None:
        """
        Initializes a Selector object.

        Args:
            target (Target, optional): The target type. Defaults to Target.S (self).
        """
        self.target = target
        self.arguments = []

    def _args(self, **args):
        for key, value in args.items():
            if not value is None and {key: value} not in self.arguments:
                self.arguments.append({key: value})
        return self

    def type(self, *types: str | Identifier):
        """Sets the type of the target."""
        for type in types:
            self._args(type=type)
        return self

    def name(self, name: str):
        """Sets the name of the target.

        Args:
            name (str): The name of the target.

        """
        self._args(name=f'"{name}"')
        return self

    def family(self, family: str):
        """Sets the family of the target.

        Args:
            family (str): The family of the target.

        """
        self._args(family=family)
        return self

    def count(self, count: int):
        """Sets the count of the target.

        Args:
            count (int): The number of targets to select.

        """
        self._args(c=count)
        return self

    def coordinates(self, *, x: coordinate = None, y: coordinate = None, z: coordinate = None):
        """Sets the coordinates of the target.

        Args:
            x (coordinate, optional): The x coordinate. Defaults to None.
            y (coordinate, optional): The y coordinate. Defaults to None.
            z (coordinate, optional): The z coordinate. Defaults to None.

        """
        self._args(x=x, y=y, z=z)
        return self

    def distance(self, *, r: coordinate = None, rm: coordinate = None):
        """Sets the distance of the target.

        Args:
            r (coordinate, optional): The maximum distance. Defaults to None.
            rm (coordinate, optional): The minimum distance. Defaults to None.

        """
        self._args(r=r, rm=rm)
        return self

    def volume(self, *, dx: float = None, dy: float = None, dz: float = None):
        """Sets the volume of the target.

        Args:
            dx (float, optional): The x volume. Defaults to None.
            dy (float, optional): The y volume. Defaults to None.
            dz (float, optional): The z volume. Defaults to None.

        """
        self._args(dx=dx, dy=dy, dz=dz)
        return self

    def scores(self, **scores):
        """Sets the scores of the target.

        Example:
            >>> selector.scores(score1=1, score2=2, score3=3)
        """
        score_values = {}
        for score, value in scores.items():
            score_values.update({score: value})
        self._args(scores=score_values)
        return self

    def tag(self, *tags: str):
        """Sets the tags of the target.

        Example:
            >>> selector.tag("tag1", "!tag2")
        """
        for tag in tags:
            self._args(tag=tag)
        return self

    def rotation(
        self,
        *,
        ry: float = None,
        rym: float = None,
        rx: float = None,
        rxm: float = None,
    ):
        """Sets the rotation of the target.

        Args:
            ry (float, optional): The maximum yaw. Defaults to None.
            rym (float, optional): The minimum yaw. Defaults to None.
            rx (float, optional): The maximum pitch. Defaults to None.
            rxm (float, optional): The minimum pitch. Defaults to None.

        """
        self._args(
            ry=normalize_180(round(ry, 2)) if not ry is None else ry,
            rym=normalize_180(round(rym, 2)) if not rym is None else rym,
            rx=clamp(round(rx, 2), -90, 90) if not rx is None else rx,
            rxm=clamp(round(rxm, 2), -90, 90) if not rxm is None else rxm,
        )
        return self

    def has_permission(self, *, camera: bool = None, movement: bool = None):
        """Selects targets that have the specified permissions.

        Args:
            camera (bool, optional): Defaults to None.
            movement (bool, optional): Defaults to None.

        """
        permission = {}
        if not camera is None:
            permission.update({"camera": "enabled" if camera else "disabled"})
        if not movement is None:
            permission.update({"movement": "enabled" if movement else "disabled"})

        self._args(haspermission=permission)
        return self

    def has_item(
        self,
        *,
        item: str | Identifier,
        data: int = -1,
        quantity: int = None,
        location: Slots = None,
        slot: int = None,
    ):
        """Selects targets that have the specified item.

        Args:
            item (str | Identifier): The item to check for.
            data (int, optional): The data value of the item. Defaults to -1.
            quantity (int, optional): The quantity of the item. Defaults to None.
            location (Slots, optional): The location of the item. Defaults to None.
            slot (int, optional): The slot of the item. Defaults to None.

        """
        test_item = {
            "item": item,
            "data": data if data != -1 else None,
            "quantity": quantity,
            "location": location,
            "slot": slot,
        }
        self._args(hasitem=test_item)
        return self

    def gamemode(self, gamemode: Gamemodes):
        """Selects targets that have the specified gamemode.

        Args:
            gamemode (Gamemodes): The gamemode to check for.

        """
        self._args(m=gamemode)
        return self

    def has_property(self, **properties):
        """Selects targets that have the specified property.

        Args:
            property (str): The property to check for.
            value (str): The value of the property.

        """
        from anvil import ANVIL

        property_values = {}
        for property, value in properties.items():
            property_values.update({f"{ANVIL.config.NAMESPACE}:{property}": json.dumps(value)})
        self._args(has_property=property_values)
        return self

    def __str__(self) -> str:
        """Returns the target selector as a string."""
        if len(self.arguments) > 0:
            args = []
            for i in self.arguments:
                for key, value in i.items():
                    values = value
                    if type(value) is dict:
                        values = f"{{{', '.join(f'{k} = {v}' for k, v in value.items() if not v is None)}}}"
                    args.append(f"{key} = {values}")

            self.target = f"{self.target} [{', '.join(args)}]"
        return self.target


class RawTextConstructor:
    """
    A class that constructs raw text with various possible styling and structures.

    Attributes:
        _raw_text (list): The raw text being constructed, stored as a list of dictionaries representing different parts.
    """

    def __init__(self):
        """Initializes a new RawTextConstructor instance."""
        self._raw_text = []

    def style(self, *styles: Style):
        """
        Applies the provided styles to the raw text.

        Args:
            styles (Style): The styles to be applied.

        Returns:
            self: The instance of the current RawTextConstructor.
        """
        self._raw_text.append({"text": "".join(map(lambda a: a.value, styles))})
        return self

    def text(self, text):
        """
        Appends the given text to the raw text.

        Args:
            text (str): The text to be appended.

        Returns:
            self: The instance of the current RawTextConstructor.
        """
        self._raw_text.append({"text": str(text)})
        return self

    def translate(self, text):
        """
        Localizes the given text and appends it to the raw text.

        Args:
            text (str): The text to be localized and appended.

        Returns:
            self: The instance of the current RawTextConstructor.
        """
        from anvil import ANVIL

        self.id = ANVIL.definitions.get_new_lang
        ANVIL.definitions.register_lang(self.id, text)
        self._raw_text.append({"translate": self.id, "with": ["\n"]})
        return self

    def score(self, objective, target: Selector | Target | str):
        """
        Appends the score of the given target under the given objective to the raw text.

        Args:
            objective (str): The objective under which the target's score is tracked.
            target (Selector | Target | str): The target whose score is to be appended.

        Returns:
            self: The instance of the current RawTextConstructor.
        """
        t = target.value if isinstance(target, Target) else target

        self._raw_text.append({"score": {"name": t, "objective": objective}})
        return self

    def selector(self, target: Selector | Target | str):
        """
        Appends the selector of the given target to the raw text.

        Args:
            target (Selector | Target | str): The target whose selector is to be appended.

        Returns:
            self: The instance of the current RawTextConstructor.
        """
        t = target.value if isinstance(target, Target) else target

        self._raw_text.append({"selector": t})
        return self

    def __str__(self) -> str:
        """
        Converts the RawTextConstructor instance to a string.

        Returns:
            str: A string representation of the raw text.
        """
        return f'{{"rawtext":{json.dumps(self._raw_text, ensure_ascii=False)}}}'


class Population(StrEnum):
    """
    Enumeration for the different types of in-game populations.
    """

    Animal = "animal"
    UnderwaterAnimal = "underwater_animal"
    Monster = "monster"
    Ambient = "ambient"


class PlacementDirectionTrait(StrEnum):
    CardinalDirection = "minecraft:cardinal_direction"  # North, South, East, West
    FacingDirection = "minecraft:facing_direction"  # Up, Down, North, South, East, West


class PlacementPositionTrait(StrEnum):
    BlockFace = "minecraft:block_face"  # Up, Down, North, South, East, West
    VerticalHalf = "minecraft:vertical_half"  # Top, Bottom


class CardinalDirectionsTrait(StrEnum):
    SOUTH = "south"
    WEST = "west"
    NORTH = "north"
    EAST = "east"


class FacingDirectionsTrait(StrEnum):
    Up = "up"
    Down = "down"
    SOUTH = "south"
    WEST = "west"
    NORTH = "north"
    EAST = "east"


class BlockFacesTrait(StrEnum):
    Up = "up"
    Down = "down"
    SOUTH = "south"
    WEST = "west"
    NORTH = "north"
    EAST = "east"


class VerticalHalfTrait(StrEnum):
    TOP = "top"
    BOTTOM = "bottom"


class BlockFaces(StrEnum):
    """
    Enumeration representing the different faces of a block in Minecraft.
    """

    Up = "up"
    Down = "down"
    North = "north"
    South = "south"
    East = "east"
    West = "west"
    Side = "side"
    All = "all"


class BlockMaterial(StrEnum):
    """
    Enumeration representing the different types of rendering methods a block can use in Minecraft.
    """

    Opaque = "opaque"
    DoubleSided = "double_sided"
    Blend = "blend"
    AlphaTest = "alpha_test"
    AlphaTestSingleSided = "alpha_test_single_sided"


class ItemCategory(StrEnum):
    """
    Enumeration representing the categories of blocks and items that can be used in Minecraft.
    """

    Construction = "construction"
    Equipment = "equipment"
    Items = "items"
    Nature = "nature"
    none = "none"


class ItemGroups(StrEnum):
    Anvil = "itemGroup.name.anvil"
    Arrow = "itemGroup.name.arrow"
    Axe = "itemGroup.name.axe"
    Banner = "itemGroup.name.banner"
    BannerPattern = "itemGroup.name.banner_pattern"
    Bed = "itemGroup.name.bed"
    Boat = "itemGroup.name.boat"
    Boots = "itemGroup.name.boots"
    Buttons = "itemGroup.name.buttons"
    Candles = "itemGroup.name.candles"
    Chalkboard = "itemGroup.name.chalkboard"
    Chest = "itemGroup.name.chest"
    ChestBoat = "itemGroup.name.chestboat"
    Chestplate = "itemGroup.name.chestplate"
    Concrete = "itemGroup.name.concrete"
    ConcretePowder = "itemGroup.name.concretePowder"
    CookedFood = "itemGroup.name.cookedFood"
    Copper = "itemGroup.name.copper"
    Coral = "itemGroup.name.coral"
    CoralDecorations = "itemGroup.name.coral_decorations"
    Crop = "itemGroup.name.crop"
    Door = "itemGroup.name.door"
    Dye = "itemGroup.name.dye"
    EnchantedBook = "itemGroup.name.enchantedBook"
    Fence = "itemGroup.name.fence"
    FenceGate = "itemGroup.name.fenceGate"
    Firework = "itemGroup.name.firework"
    FireworkStars = "itemGroup.name.fireworkStars"
    Flower = "itemGroup.name.flower"
    Glass = "itemGroup.name.glass"
    GlassPane = "itemGroup.name.glassPane"
    GlazedTerracotta = "itemGroup.name.glazedTerracotta"
    GoatHorn = "itemGroup.name.goatHorn"
    Grass = "itemGroup.name.grass"
    HangingSign = "itemGroup.name.hanging_sign"
    Helmet = "itemGroup.name.helmet"
    Hoe = "itemGroup.name.hoe"
    HorseArmor = "itemGroup.name.horseArmor"
    Leaves = "itemGroup.name.leaves"
    Leggings = "itemGroup.name.leggings"
    LingeringPotion = "itemGroup.name.lingeringPotion"
    Log = "itemGroup.name.log"
    Minecart = "itemGroup.name.minecart"
    MiscFood = "itemGroup.name.miscFood"
    MobEgg = "itemGroup.name.mobEgg"
    MonsterStoneEgg = "itemGroup.name.monsterStoneEgg"
    Mushroom = "itemGroup.name.mushroom"
    NetherWartBlock = "itemGroup.name.netherWartBlock"
    Ore = "itemGroup.name.ore"
    Permission = "itemGroup.name.permission"
    Pickaxe = "itemGroup.name.pickaxe"
    Planks = "itemGroup.name.planks"
    Potion = "itemGroup.name.potion"
    PotterySherds = "itemGroup.name.potterySherds"
    PressurePlate = "itemGroup.name.pressurePlate"
    Rail = "itemGroup.name.rail"
    RawFood = "itemGroup.name.rawFood"
    Record = "itemGroup.name.record"
    Sandstone = "itemGroup.name.sandstone"
    Sapling = "itemGroup.name.sapling"
    Sculk = "itemGroup.name.sculk"
    Seed = "itemGroup.name.seed"
    Shovel = "itemGroup.name.shovel"
    ShulkerBox = "itemGroup.name.shulkerBox"
    Sign = "itemGroup.name.sign"
    Skull = "itemGroup.name.skull"
    Slab = "itemGroup.name.slab"
    SmithingTemplates = "itemGroup.name.smithing_templates"
    SplashPotion = "itemGroup.name.splashPotion"
    StainedClay = "itemGroup.name.stainedClay"
    Stairs = "itemGroup.name.stairs"
    Stone = "itemGroup.name.stone"
    StoneBrick = "itemGroup.name.stoneBrick"
    Sword = "itemGroup.name.sword"
    Trapdoor = "itemGroup.name.trapdoor"
    Walls = "itemGroup.name.walls"
    Wood = "itemGroup.name.wood"
    Wool = "itemGroup.name.wool"
    WoolCarpet = "itemGroup.name.woolCarpet"
    none = "None"


class WeatherSet(StrEnum):
    """
    Enumeration representing the different types of weather that can be set in Minecraft.
    """

    Clear = "clear"
    Rain = "rain"
    Thunder = "thunder"


class FilterSubject(StrEnum):
    """
    Enumeration representing the different subjects that can be used in filters in Minecraft.
    """

    Block = "block"
    Damager = "damager"
    Other = "other"
    Parent = "parent"
    Player = "player"
    Self = "self"
    Target = "target"


class FilterOperation(StrEnum):
    """
    Enumeration representing the different operations that can be used in filters in Minecraft.
    """

    Less = "<"
    LessEqual = "<="
    Greater = ">"
    GreaterEqual = ">="
    Equals = "equals"
    Not = "not"


class FilterEquipmentDomain(StrEnum):
    """
    Enumeration representing the different equipment domains that can be used in filters in Minecraft.
    """

    Any = "any"
    Armor = "armor"
    Feet = "feet"
    Hand = "hand"
    Head = "head"
    Inventory = "inventory"
    Leg = "leg"
    Torso = "torso"


class Biomes(StrEnum):
    BEACH = "beach"
    DESERT = "desert"
    EXTREME_HILLS = "extreme_hills"
    FLAT = "flat"
    FOREST = "forest"
    ICE = "ice"
    JUNGLE = "jungle"
    MESA = "mesa"
    MUSHROOM_ISLAND = "mushroom_island"
    OCEAN = "ocean"
    PLAIN = "plain"
    RIVER = "river"
    SAVANNA = "savanna"
    STONE_BEACH = "stone_beach"
    SWAMP = "swamp"
    TAIGA = "taiga"
    THE_END = "the_end"
    THE_NETHER = "the_nether"


class Vibrations(StrEnum):
    EntityInteract = "entity_interact"
    EntityAct = "entity_act"
    Shear = "shear"
    none = "none"


class ControlFlags(StrEnum):
    Move = "move"
    Look = "look"


class MaterialStates(StrEnum):
    """
    Enumeration representing the different states a material can be in Minecraft.
    """

    EnableStencilTest = "EnableStencilTest"
    StencilWrite = "StencilWrite"
    InvertCulling = "InvertCulling"
    DisableCulling = "DisableCulling"
    DisableDepthWrite = "DisableDepthWrite"
    Blending = "Blending"


class MaterialDefinitions(StrEnum):
    """
    Enumeration representing the different definitions that can be set for a material in Minecraft.
    """

    Fancy = "FANCY"
    USE_OVERLAY = "USE_OVERLAY"
    USE_COLOR_MASK = "USE_COLOR_MASK"
    MULTI_COLOR_TINT = "MULTI_COLOR_TINT"
    COLOR_BASED = "COLOR_BASED"
    USE_UV_ANIM = "USE_UV_ANIM"
    TINTED = "TINTED"
    USE_COLOR_BLEND = "USE_COLOR_BLEND"
    MULTIPLICATIVE_TINT = "MULTIPLICATIVE_TINT"
    MULTIPLICATIVE_TINT_COLOR = "MULTIPLICATIVE_TINT_COLOR"


class MaterialFunc(StrEnum):
    """
    Enumeration representing the different functions that can be set for a material in Minecraft.
    """

    Always = "Always"
    Equal = "Equal"
    NotEqual = "NotEqual"
    Less = "Less"
    Greater = "Greater"
    GreaterEqual = "GreaterEqual"
    LessEqual = "LessEqual"


class MaterialOperation(StrEnum):
    """
    Enumeration representing the different operations that can be set for a material in Minecraft.
    """

    Keep = "Keep"
    Replace = "Replace"


class SoundCategory(StrEnum):
    """
    Enumeration for the different categories of sounds in the game.
    """

    Ambient = "ambient"
    Block = "block"
    Player = "player"
    Neutral = "neutral"
    Hostile = "hostile"
    Music = "music"
    Record = "record"
    UI = "ui"


class EntitySoundEvent(StrEnum):
    Ambient = "ambient"
    Hurt = "hurt"
    Death = "death"
    Takeoff = "takeoff"
    AmbientTame = "ambient.tame"
    Purr = "purr"
    Purreow = "purreow"
    Eat = "eat"
    Step = "step"
    Plop = "plop"
    Fuse = "fuse"
    Breathe = "breathe"
    Attack = "attack"
    Splash = "splash"
    Swim = "swim"
    AmbientInWater = "ambient.in.water"
    HurtInWater = "hurt.in.water"
    DeathInWater = "death.in.water"
    Jump = "jump"
    Mad = "mad"
    Stare = "stare"
    Flap = "flap"
    Fizz = "fizz"
    Sniff = "sniff"
    Screech = "screech"
    Sleep = "sleep"
    Spit = "spit"
    Scream = "scream"
    Warn = "warn"
    Shoot = "shoot"
    GuardianFlop = "guardian.flop"
    Flop = "flop"
    CastSpell = "cast.spell"
    PrepareAttack = "prepare.attack"
    PrepareSummon = "prepare.summon"
    PrepareWololo = "prepare.wololo"
    AmbientInRaid = "ambient.in.raid"
    Celebrate = "celebrate"
    Fang = "fang"
    Charge = "charge"
    Armor = "armor"
    AddChest = "add.chest"
    Saddle = "saddle"
    Land = "land"
    Throw = "throw"
    AttackStrong = "attack.strong"
    Roar = "roar"
    Stun = "stun"
    Thunder = "thunder"
    Explode = "explode"
    Fly = "fly"
    ImitateBlaze = "imitate.blaze"
    ImitateCaveSpider = "imitate.cave_spider"
    ImitateCreeper = "imitate.creeper"
    ImitateElderGuardian = "imitate.elder_guardian"
    ImitateEnderDragon = "imitate.ender_dragon"
    ImitateEnderman = "imitate.enderman"
    ImitateEndermite = "imitate.endermite"
    ImitateEvocationIllager = "imitate.evocation_illager"
    ImitateGhast = "imitate.ghast"
    ImitateHusk = "imitate.husk"
    ImitateIllusionIllager = "imitate.illusion_illager"
    ImitateMagmaCube = "imitate.magma_cube"
    ImitatePolarBear = "imitate.polar_bear"
    ImitatePanda = "imitate.panda"
    ImitateShulker = "imitate.shulker"
    ImitateSilverfish = "imitate.silverfish"
    ImitateSkeleton = "imitate.skeleton"
    ImitateSlime = "imitate.slime"
    ImitateSpider = "imitate.spider"
    ImitateStray = "imitate.stray"
    ImitateVex = "imitate.vex"
    ImitateVindicationIllager = "imitate.vindication_illager"
    ImitateWitch = "imitate.witch"
    ImitateWither = "imitate.wither"
    ImitateWitherSkeleton = "imitate.wither_skeleton"
    ImitateWolf = "imitate.wolf"
    ImitateZombie = "imitate.zombie"
    ImitateDrowned = "imitate.drowned"
    ImitateZombiePigman = "imitate.zombie_pigman"
    ImitateZombieVillager = "imitate.zombie_villager"
    Swoop = "swoop"
    Boost = "boost"
    DeathToZombie = "death.to.zombie"
    AttackNodamage = "attack.nodamage"
    ElderguardianCurse = "elderguardian.curse"
    AmbientBaby = "ambient.baby"
    MobWarning = "mob.warning"
    AmbientAggressive = "ambient.aggressive"
    AmbientWorried = "ambient.worried"
    Presneeze = "presneeze"
    Sneeze = "sneeze"
    CantBreed = "cant_breed"
    ShulkerOpen = "shulker.open"
    ShulkerClose = "shulker.close"
    HurtBaby = "hurt.baby"
    DeathBaby = "death.baby"
    StepBaby = "step.baby"
    Born = "born"
    SquishBig = "squish.big"
    SquishSmall = "squish.small"
    Haggle = "haggle"
    HaggleYes = "haggle.yes"
    HaggleNo = "haggle.no"
    Disappeared = "disappeared"
    Drink = "drink"
    Reappeared = "reappeared"
    DeathMinVolume = "death.min.volume"
    DeathMidVolume = "death.mid.volume"
    Spawn = "spawn"
    BreakBlock = "break.block"
    Shake = "shake"
    Growl = "growl"
    Whine = "whine"
    Pant = "pant"


class MusicCategory(StrEnum):
    """
    Enumeration for the different categories of music in the game.
    """

    Creative = "creative"
    Credits = "credits"
    Crimson_forest = "crimson_forest"
    Dripstone_caves = "dripstone_caves"
    End = "end"
    Endboss = "endboss"
    Frozen_peaks = "frozen_peaks"
    Game = "game"
    Grove = "grove"
    Hell = "hell"
    Jagged_peaks = "jagged_peaks"
    Lush_caves = "lush_caves"
    Meadow = "meadow"
    Menu = "menu"
    Nether = "nether"
    Snowy_slopes = "snowy_slopes"
    Soulsand_valley = "soulsand_valley"
    Stony_peaks = "stony_peaks"
    Water = "water"


class SmeltingTags(StrEnum):
    FURNACE = "furnace"
    BLAST_FURNACE = "blast_furnace"
    SMOKER = "smoker"
    CAMPFIRE = "campfire"
    SOUL_CAMPFIRE = "soul_campfire"


class ContainerType(StrEnum):
    Horse = "horse"
    MinecartChest = "minecart_chest"
    ChestBoat = "chest_boat"
    MinecartHopper = "minecart_hopper"
    Inventory = "inventory"
    Container = "container"
    Hopper = "hopper"


class EnchantsSlots(StrEnum):
    ArmorHead = "armor_head"
    ArmorTorse = "armor_torso"
    ArmorLegs = "armor_legs"
    ArmorFeet = "armor_feet"


class HudElement(StrEnum):
    Hunger = "hunger"
    All = "all"
    Paperdoll = "paperdoll"
    Armor = "armor"
    Tooltips = "tooltips"
    TouchControls = "touch_controls"
    Crosshair = "crosshair"
    Hotbar = "hotbar"
    Health = "health"
    ProgressBar = "progress_bar"
    AirBubbles = "air_bubbles"
    HorseHealth = "horse_health"
    StatusEffects = "status_effects"
    ItemText = "item_text"


class HudVisibility(StrEnum):
    Hide = "hide"
    Reset = "reset"


class ItemVanillaTags(StrEnum):
    Arrow = "minecraft:arrow"
    Banner = "minecraft:banner"
    Boat = "minecraft:boat"
    Boats = "minecraft:boats"
    BookshelfBooks = "minecraft:bookshelf_books"
    ChainmailTier = "minecraft:chainmail_tier"
    Coals = "minecraft:coals"
    CrimsonStems = "minecraft:crimson_stems"
    DecoratedPotSherds = "minecraft:decorated_pot_sherds"
    DiamondTier = "minecraft:diamond_tier"
    Digger = "minecraft:digger"
    Door = "minecraft:door"
    GoldenTier = "minecraft:golden_tier"
    HangingActor = "minecraft:hanging_actor"
    HangingSign = "minecraft:hanging_sign"
    HorseArmor = "minecraft:horse_armor"
    IronTier = "minecraft:iron_tier"
    IsArmor = "minecraft:is_armor"
    IsAxe = "minecraft:is_axe"
    IsCooked = "minecraft:is_cooked"
    IsFish = "minecraft:is_fish"
    IsFood = "minecraft:is_food"
    IsHoe = "minecraft:is_hoe"
    IsMeat = "minecraft:is_meat"
    IsMinecart = "minecraft:is_minecart"
    IsPickaxe = "minecraft:is_pickaxe"
    IsShovel = "minecraft:is_shovel"
    IsSword = "minecraft:is_sword"
    IsTool = "minecraft:is_tool"
    IsTrident = "minecraft:is_trident"
    LeatherTier = "minecraft:leather_tier"
    LecternBooks = "minecraft:lectern_books"
    Logs = "minecraft:logs"
    LogsThatBurn = "minecraft:logs_that_burn"
    MangroveLogs = "minecraft:mangrove_logs"
    MusicDisc = "minecraft:music_disc"
    NetheriteTier = "minecraft:netherite_tier"
    Planks = "minecraft:planks"
    Sand = "minecraft:sand"
    Sign = "minecraft:sign"
    SoulFireBaseBlocks = "minecraft:soul_fire_base_blocks"
    SpawnEgg = "minecraft:spawn_egg"
    StoneBricks = "minecraft:stone_bricks"
    StoneCraftingMaterials = "minecraft:stone_crafting_materials"
    StoneTier = "minecraft:stone_tier"
    StoneToolMaterials = "minecraft:stone_tool_materials"
    TransformMaterials = "minecraft:transform_materials"
    TransformTemplates = "minecraft:transform_templates"
    TransformableItems = "minecraft:transformable_items"
    TrimTemplates = "minecraft:trim_templates"
    TrimmableArmors = "minecraft:trimmable_armors"
    VibrationDamper = "minecraft:vibration_damper"
    WarpedStems = "minecraft:warped_stems"
    WoodenSlabs = "minecraft:wooden_slabs"
    WoodenTier = "minecraft:wooden_tier"
    Wool = "minecraft:wool"


class BlockVanillaTags(StrEnum):
    Wood = "wood"
    Pumpkin = "pumpkin"
    Plant = "plant"
    Stone = "stone"
    Metal = "metal"
    DiamondPickDiggable = "diamond_pick_diggable"
    GoldPickDiggable = "gold_pick_diggable"
    IronPickDiggable = "iron_pick_diggable"
    StonePickDiggable = "stone_pick_diggable"
    WoodPickDiggable = "wood_pick_diggable"
    Dirt = "dirt"
    Sand = "sand"
    Gravel = "gravel"
    Grass = "grass"
    Snow = "snow"
    Rail = "rail"
    Water = "water"
    MobSpawner = "mob_spawner"
    LushPlantsReplaceable = "lush_plants_replaceable"
    AzaleaLogReplaceable = "azalea_log_replaceable"
    NotFeatureReplaceable = "not_feature_replaceable"
    TextSign = "text_sign"
    Crop = "minecraft:crop"
    FertilizeArea = "fertilize_area"

    DiamondTierDestructible = "minecraft:diamond_tier_destructible"
    IronTierDestructible = "minecraft:iron_tier_destructible"
    IsAxeItemDestructible = "minecraft:is_axe_item_destructible"
    IsHoeItemDestructible = "minecraft:is_hoe_item_destructible"
    IsMaceItemDestructible = "minecraft:is_mace_item_destructible"
    IsPickaxeItemDestructible = "minecraft:is_pickaxe_item_destructible"
    IsShearsItemDestructible = "minecraft:is_shears_item_destructible"
    IsShovelItemDestructible = "minecraft:is_shovel_item_destructible"
    IsSwordItemDestructible = "minecraft:is_sword_item_destructible"
    NetheriteTierDestructible = "minecraft:netherite_tier_destructible"
    StoneTierDestructible = "minecraft:stone_tier_destructible"


class ItemRarity(StrEnum):
    Common = "common"
    Uncommon = "uncommon"
    Rare = "rare"
    Epic = "epic"


class PotionId(StrEnum):
    Water = "water"
    Mundane = "mundane"
    LongMundane = "long_mundane"
    Thick = "thick"
    Awkward = "awkward"
    NightVision = "nightvision"
    LongNightVision = "long_nightvision"
    Invisibility = "invisibility"
    LongInvisibility = "long_invisibility"
    Leaping = "leaping"
    LongLeaping = "long_leaping"
    StrongLeaping = "strong_leaping"
    FireResistance = "fire_resistance"
    LongFireResistance = "long_fire_resistance"
    Swiftness = "swiftness"
    LongSwiftness = "long_swiftness"
    StrongSwiftness = "strong_swiftness"
    Slowness = "slowness"
    LongSlowness = "long_slowness"
    StrongSlowness = "strong_slowness"
    WaterBreathing = "water_breathing"
    LongWaterBreathing = "long_water_breathing"
    Healing = "healing"
    StrongHealing = "strong_healing"
    Harming = "harming"
    StrongHarming = "strong_harming"
    Poison = "poison"
    LongPoison = "long_poison"
    StrongPoison = "strong_poison"
    Regeneration = "regeneration"
    LongRegeneration = "long_regeneration"
    StrongRegeneration = "strong_regeneration"
    Strength = "strength"
    LongStrength = "long_strength"
    StrongStrength = "strong_strength"
    Weakness = "weakness"
    LongWeakness = "long_weakness"
    Wither = "wither"
    TurtleMaster = "turtle_master"
    LongTurtleMaster = "long_turtle_master"
    StrongTurtleMaster = "strong_turtle_master"
    SlowFalling = "slow_falling"
    LongSlowFalling = "long_slow_falling"
    WindCharged = "wind_charged"
    Weaving = "weaving"
    Oozing = "oozing"
    Infested = "infested"


class ExplosionParticleEffect(StrEnum):
    Explosion = "explosion"
    WindBurst = "wind_burst"
    BreezeWindBurst = "breeze_wind_burst"


class DamageSensorDamage(StrEnum):
    Yes = "yes"
    No = "no"
    NoButSideEffectsApply = "no_but_side_effects_apply"

class LineOfSightObstructionType(StrEnum):
    Outline = "outline"
    Collision = "collision"
    CollisionForCamera = "collision_for_camera"

class LookAtLocation(StrEnum):
    Head = "head"
    Body = "body"
    Feet = "feet"

class LootedAtSetTarget(StrEnum):
    Never = "never"
    OnceAndStopScanning = "once_and_stop_scanning"
    OnceAndKeepScanning = "once_and_keep_scanning"


import json
from enum import StrEnum
from math import inf

from anvil import CONFIG
from anvil.api.blocks import Block
from anvil.api.enums import (Anchor, CameraEasing, CameraPresets,
                             CameraShakeType, CloneMode, DamageCause,
                             Dimension, Effects, FillMode, Gamemodes,
                             HudElement, HudVisibility, InputPermissions,
                             MaskMode, MusicRepeatMode, RawTextConstructor,
                             ScoreboardOperation, ScoreboardOperator, Selector,
                             Slots, Target, TimeSpec)
from anvil.api.types import (Color, Identifier, Seconds, coordinate,
                             coordinates, event, position, rotation, tick)
from anvil.api.vanilla import Blocks
from anvil.lib.lib import clamp, normalize_180


class Command:
    def __init__(self, prefix: str, *commands) -> None:
        self._prefix = prefix
        self._components = {}
        self._command = f"{self._prefix} {' '.join([str(cmd) for cmd in commands])}"

    def _new_cmd(self, *commands):
        self._command = f"{self._prefix} {' '.join([str(cmd) for cmd in commands])}"
        return self

    def _append_cmd(self, *commands):
        self._command += f" {' '.join([str(cmd) for cmd in commands])}"
        return self

    def _add_nbt_component(self, component: dict):
        self._components.update(component)
        return self

    def __str__(self):
        self._command = self._command.lstrip("/")
        if len(self._components) > 0:
            self._append_cmd(json.dumps(self._components))
            self._components = {}
        return self._command


class ItemComponents(Command):
    def can_place_on(self, *blocks):
        self._add_nbt_component({"minecraft:can_place_on": {"blocks": blocks}})
        return self

    def can_destroy(self, *blocks):
        self._add_nbt_component({"minecraft:can_destroy": {"blocks": blocks}})
        return self

    def item_lock(self, *, lock_in_slot: bool = False, lock_in_inventory: bool = False):
        if lock_in_slot or lock_in_inventory:
            self._add_nbt_component({"minecraft:item_lock": {"mode": "lock_in_slot" if lock_in_slot else "lock_in_inventory"}})
        return self

    @property
    def keep_on_death(self):
        self._add_nbt_component({"minecraft:keep_on_death": {}})
        return self


# special
class _RawText(RawTextConstructor):
    def __init__(self, texttype: str, target):
        self._type = texttype
        self._target = target
        if self._type in ["actionbar", "subtitle", "title"]:
            self._command = f"titleraw {self._target} {self._type} "
        elif self._type == "tellraw":
            self._command = f"tellraw {self._target} "
        super().__init__()

    def __str__(self) -> str:
        return self._command + super().__str__()


class TitleRaw(Command):
    def __init__(self) -> None:
        super().__init__("titleraw")

    def title(self, target: Selector | Target):
        return _RawText("title", target)

    def subtitle(self, target: Selector | Target):
        return _RawText("subtitle", target)

    def actionbar(self, target: Selector | Target):
        return _RawText("actionbar", target)

    def times(self, target: Selector | Target, fade_in: int, stay: int, fade_out: int):
        self._append_cmd(target, "times", fade_in, stay, fade_out)
        return self


class Tellraw(Command):
    def __init__(self, target: str):
        self._target = target

    @property
    def text(self):
        return _RawText("tellraw", self._target)


class Execute(Command):
    class _ExecuteIfUnless:
        def __init__(self, parent: "Execute", condition: str) -> None:
            self._parent = parent
            self._condition = condition

        def Entity(self, target: Selector):
            self._parent._append_cmd(self._condition, "entity", target)
            return self._parent

        def Block(self, block_position: coordinates, tile: Block | str, **properties):
            name = (
                tile.identifier
                if isinstance(tile, (Blocks._MinecraftBlock, Block))
                else tile if isinstance(tile, str) else CONFIG.Logger.unsupported_block_type(tile)
            )
            states = (
                tile.states
                if isinstance(tile, Blocks._MinecraftBlock)
                else "" if isinstance(tile, (str, Block)) else CONFIG.Logger.unsupported_block_type(tile)
            )

            for k, v in properties.items():
                states.append(f'"{k}" : "{v}"')

            self._parent._append_cmd(self._condition, "block", *block_position, name, f'[{", ".join(states)}]')
            return self._parent

        def Blocks(self, begin: coordinates, end: coordinates, destination: coordinates, masked: bool = False):
            self._parent._append_cmd(self._condition, "blocks", *begin, *end, *destination, masked if masked == True else "")
            return self._parent

        def ScoreMatches(self, target: str, objective: str, matches: range):
            self._parent._append_cmd(
                self._condition, "score", target, objective, "matches", round(matches) if isinstance(matches, float) else matches
            )
            return self._parent

        def Score(self, target: str, target_objective: str, operator: ScoreboardOperator, source: str, source_objective: str):
            self._parent._append_cmd(self._condition, "score", target, target_objective, operator, source, source_objective)
            return self._parent

    def __init__(self) -> None:
        super().__init__("execute")

    def As(self, target: str):
        super()._append_cmd("as", target)
        return self

    def At(self, target: str):
        super()._append_cmd("at", target)
        return self

    def In(self, dimension: Dimension = Dimension.Overworld):
        super()._append_cmd("in", dimension)
        return self

    def Positioned(self, poistion: coordinates):
        super()._append_cmd("positioned", " ".join(map(str, poistion)))
        return self

    def PositionedAs(self, target: str):
        super()._append_cmd("positioned", "as", target)
        return self

    def Align(self, axes: str = "xyz"):
        super()._append_cmd("align", "".join(set(axes)))
        return self

    def Anchored(self, anchored: Anchor = Anchor.Feet):
        super()._append_cmd("anchored", anchored)
        return self

    def Rotated(self, yaw: float | str = "~", pitch: float | str = "~"):
        super()._append_cmd("rotated", yaw, pitch)
        return self

    def RotatedAs(self, target: str):
        super()._append_cmd("rotated", "as", target)
        return self

    def Facing(self, poistion: coordinates):
        super()._append_cmd("facing", " ".join(map(str, poistion)))
        return self

    def FacingEntity(self, target: str, anchor: Anchor = Anchor.Feet):
        super()._append_cmd("facing", "entity", target, anchor)
        return self

    @property
    def If(self):
        return self._ExecuteIfUnless(self, "if")

    @property
    def Unless(self):
        return self._ExecuteIfUnless(self, "unless")

    def run(self, command: Command | str):
        super()._append_cmd("run", command)
        return self


class AlwaysDay(Command):
    def __init__(self, lock: bool = True):
        super().__init__("alwaysday", lock)


class CameraShake(Command):
    def __init__(self) -> None:
        super().__init__("camerashake")

    def add(self, target: str, intensity: float, seconds: float, shakeType: CameraShakeType):
        super()._append_cmd("add", target, intensity, seconds, shakeType)
        return self

    def stop(self, target: str):
        super()._append_cmd("stop", target)
        return self


class Summon(Command):
    def __init__(self, entity: Identifier, coordinates: coordinates = ("~", "~", "~")):
        self.argument = {
            "coordinates": (str(c) for c in coordinates),
            "lookAtEntity": None,
            "lookAtPosition": None,
            "rotation": None,
            "spawnEvent": None,
            "nameTag": None,
        }
        super().__init__("summon", entity)

    def nameTag(self, nameTag: str):
        self.argument["nameTag"] = f'"{nameTag}"'
        return self

    def spawnEvent(self, spawnEvent: str):
        self.argument["spawnEvent"] = spawnEvent
        return self

    def lookAtEntity(self, entity: Selector | Target):
        self.argument["lookAtEntity"] = entity
        return self

    def lookAtPosition(self, coordinates: coordinates):
        self.argument["lookAtPosition"] = coordinates
        return self

    def rotation(self, ry="~", rx="~"):
        self.argument["rotation"] = (str(ry), str(rx))
        return self

    def __str__(self):
        rots = sum(
            [
                x is not None
                for x in (self.argument.get("lookAtEntity"), self.argument.get("lookAtPosition"), self.argument.get("rotation"))
            ]
        )
        nameTag = self.argument.get("nameTag")
        spawnEvent = self.argument.get("spawnEvent")

        if rots > 1:
            CONFIG.Logger.multiple_rotations()

        if rots == 0 and not nameTag is None and spawnEvent is None:
            self._append_cmd(nameTag)
            self._append_cmd(" ".join(self.argument.get("coordinates")))

        else:
            if rots == 0:
                self.argument["rotation"] = ("~", "~")

            self._append_cmd(" ".join(self.argument.get("coordinates")))

            if self.argument.get("lookAtEntity") != None:
                self._append_cmd("facing", self.argument.get("lookAtEntity"))

            elif self.argument.get("lookAtPosition") != None:
                self._append_cmd("facing", " ".join(self.argument.get("lookAtPosition")))

            else:
                self._append_cmd(" ".join(self.argument.get("rotation")))

            if spawnEvent != None:
                self._append_cmd(spawnEvent)

            if nameTag != None:
                if spawnEvent == None:
                    self._append_cmd("minecraft:entity_spawned")
                self._append_cmd(nameTag)

        return super().__str__()


class XP(Command):
    def __init__(self):
        """Adds or removes player experience."""
        super().__init__("xp")

    def add(self, target: Selector | Target | str, amount: int) -> Command:
        """Adds experience to a player.

        Args:
            target (Selector | Target | str): The target to add experience to.
            amount (int): The amount of experience to add.

        Returns:
            Command: The command.
        """
        super()._new_cmd(amount, target)
        return self

    def remove(self, target: Selector | Target | str, amount: int) -> Command:
        """Removes experience from a player.

        Args:
            target (Selector | Target | str): The target to remove experience from.
            amount (int): The amount of experience to remove.

        Returns:
            Command: The command.
        """
        super()._new_cmd(f"-{amount}", target)
        return self


class Weather(Command):
    def __init__(self) -> None:
        super().__init__("weather")

    @property
    def clear(self):
        return super()._new_cmd("clear")

    def rain(self, duration: tick = inf):
        return super()._new_cmd("rain", duration if duration != inf else "")

    def thunder(self, duration: tick = inf):
        return super()._new_cmd("thunder", duration if duration != inf else "")


class Clone(Command):
    def __init__(
        self,
        begin: coordinates,
        end: coordinates,
        destination: coordinates,
        cloneMode: CloneMode = CloneMode.normal,
        maskMode: MaskMode = MaskMode.replace,
    ):
        """Clones a region of blocks."""
        super().__init__(
            "clone", " ".join(map(str, begin)), " ".join(map(str, end)), " ".join(map(str, destination)), cloneMode, maskMode
        )


class Msg(Command):
    def __init__(self, text: str, target: str):
        super().__init__("msg", target, text)


class DebugMsg(Command):
    def __init__(self, text: str) -> None:
        super().__init__("msg", "@a[tag=dev]", f"[§2Anvil§r]: {text}")


class Spawnpoint(Command):
    def __init__(self, target: str, spawnPos: coordinates):
        super().__init__("spawnpoint", target, " ".join(map(str, spawnPos)))


class Say(Command):
    def __init__(self, text: str):
        super().__init__("say", f'"{text}"')


class Gamerule(Command):
    def __init__(self) -> None:
        super().__init__("gamerule")

    def SendCommandFeedback(self, value: bool = True):
        return super()._new_cmd("sendcommandfeedback", value)

    def CommandBlockOutput(self, value: bool = True):
        return super()._new_cmd("commandblockoutput", value)

    def ShowTags(self, value: bool = True):
        return super()._new_cmd("showtags", value)

    def PlayerSleepingPercentage(self, percentage: int = 0):
        return super()._new_cmd("sendcommandfeedback", clamp(percentage, 0, 100))

    def dodaylightcycle(self, value: bool = True):
        return super()._new_cmd("dodaylightcycle", value)


class Fog(Command):
    def __init__(self) -> None:
        super().__init__("fog")

    def Push(self, target: str, fog_identifie: str, name: str):
        super()._new_cmd(target, "push", fog_identifie, f'"{name}"')
        return self

    def Pop(self, target: str, name: str):
        super()._new_cmd(target, "pop", f'"{name}"')
        return self

    def Remove(self, target: str, name: str):
        super()._new_cmd(target, "remove", f'"{name}"')
        return self


class Tag(Command):
    def __init__(self, target: str) -> None:
        super().__init__("tag")
        self._target = target

    def add(self, tag: str):
        super()._new_cmd(self._target, "add", tag)
        return self

    def remove(self, tag: str):
        super()._new_cmd(self._target, "remove", tag)
        return self


class Clear(Command):
    def __init__(self, target: Target, itemname: str = "", date: int = -1, max_count: int = -1) -> None:
        super().__init__("clear", target, itemname, date if not date == -1 else "", max_count if not max_count == -1 else "")


class Effect(Command):
    def __init__(self, target: Selector) -> None:
        super().__init__("effect", target)

    def clear(self, effect: Effects = None):
        self._append_cmd("clear")
        if effect:
            self._append_cmd(effect)
        return self

    def give(self, effect: Effects, seconds: int = "infinite", amplifier: int = 1, hide_particles: bool = False):
        self._append_cmd(effect, seconds, amplifier)
        if amplifier != 1:
            self._append_cmd(amplifier)
        if hide_particles:
            self._append_cmd("true")
        return self


class Gamemode(Command):
    def __init__(self, target: str, gamemode: Gamemodes):
        super().__init__("gamemode", gamemode, target)


class Teleport(Command):
    def __init__(self, target, destination: coordinates | Target | Selector, rotation: rotation = ("~", "~")) -> None:
        super().__init__("teleport", target)
        if isinstance(destination, tuple):
            self._append_cmd(
                " ".join(map(str, destination)),
                " ".join(map(str, (normalize_180(rotation[0]), rotation[1]))) if not rotation == ("~", "~") else "",
            )
        elif isinstance(destination, Target | Selector):
            self._append_cmd(destination)

    def Facing(self, poistion: coordinates):
        super()._append_cmd("facing", " ".join(map(str, poistion)))
        return self

    def FacingEntity(self, target: str):
        super()._append_cmd("facing", target)
        return self


class Event(Command):
    def __init__(self, target: Selector | Target | str, event: event) -> None:
        super().__init__("event", "entity", target, event)


class Function(Command):
    def __init__(self, path) -> None:
        super().__init__("function", path)


class Give(ItemComponents):
    def __init__(self, target: Selector | Target | str, item: str, amount: int = 1, data: int = 0) -> None:
        super().__init__("give", target, item, amount, data)


class ReplaceItem(ItemComponents):
    def __init__(self) -> None:
        super().__init__("replaceitem")

    def block(self, poistion: coordinates, slot_id: int, item_name: str, amount: int = 1, data: int = 0):
        super()._new_cmd("block", *poistion, Slots.Container, slot_id, item_name, amount, data)
        return self

    def entity(self, target: Selector | Target | str, slot: Slots, slot_id: int, item_name: str, amount: int = 1, data: int = 0):
        super()._new_cmd("entity", target, slot, slot_id, item_name, amount, data)
        return self


class Damage(ItemComponents):
    def __init__(self, target: Selector | Target | str, amount: int, cause: DamageCause) -> None:
        super().__init__("damage", target, amount, cause.value)

    def damager(self, damager: Selector | Target) -> None:
        self._append_cmd("entity", damager)
        return self


class Playsound(Command):
    def __init__(
        self,
        sound: str,
        target: Selector | Target = Target.S,
        position: position = None,
        volume: int = 1,
        pitch: int = 1,
        minimumVolume: int = 0,
    ) -> None:
        super().__init__("playsound", sound, target)

        if position != None:
            self._append_cmd(*position, volume, pitch, minimumVolume)


class InputPermission(Command):
    def __init__(self) -> None:
        super().__init__("inputpermission", "set")

    def enable(self, permission: InputPermissions, target: Selector | Target = Target.S):
        self._append_cmd(target, permission, "enabled")
        return self

    def disable(self, permission: InputPermissions, target: Selector | Target = Target.S):
        self._append_cmd(target, permission, "disabled")
        return self


class Scoreboard(Command):
    class _ScoreboardObjective:
        class _ScoreboardObjectiveDisplay:
            def __init__(self, parent: "Scoreboard") -> None:
                self.parent = parent

            def list(self, objective, ascending: bool = True):
                self.parent._append_cmd("list", objective, "ascending" if ascending else "descending")
                return self.parent

            def sidebar(self, objective: str, ascending: bool = True):
                self.parent._append_cmd("sidebar", objective, "ascending" if ascending else "descending")
                return self.parent

            def belowName(self, objective):
                self.parent._append_cmd("belowName", objective)
                return self.parent

        def __init__(self, parent: "Scoreboard") -> None:
            self.parent = parent

        def add(self, objective: str, display_name: str = None):
            cmd_args = [objective, "dummy"]
            if display_name:
                cmd_args.append(display_name)
            self.parent._append_cmd("add", *cmd_args)
            return self.parent

        def remove(self, objective: str):
            self.parent._append_cmd("remove", objective)
            return self.parent

        @property
        def list(self):
            self.parent._append_cmd("list")
            return self.parent

        @property
        def setdisplay(self):
            self.parent._append_cmd("setdisplay")
            return self._ScoreboardObjectiveDisplay(self.parent)

    class _ScoreboardPlayers:
        def __init__(self, parent: "Scoreboard") -> None:
            self.parent = parent

        def set(self, target: Selector | Target | str, objective: str, count: int):
            self.parent._append_cmd("set", target, objective, count)
            return self.parent

        def add(self, target: Selector | Target | str, objective: str, count: int):
            self.parent._append_cmd("add", target, objective, count)
            return self.parent

        def remove(self, target: Selector | Target | str, objective: str, count: int):
            self.parent._append_cmd("remove", target, objective, count)
            return self.parent

        def list(self, target: Selector | Target):
            self.parent._append_cmd("list", target)
            return self.parent

        def operation(
            self, target: Selector | Target | str, objective1: str, operation: ScoreboardOperation, source: str, objective2: str
        ):
            self.parent._append_cmd("operation", target, objective1, operation, source, objective2)
            return self.parent

        def random(self, target: Selector | Target | str, objective: str, min: int, max: int):
            self.parent._append_cmd("random", target, objective, min, max)
            return self.parent

        def reset(self, target: Selector | Target | str, objective: str):
            self.parent._append_cmd("reset", target, objective)
            return self.parent

    def __init__(self) -> None:
        super().__init__("scoreboard")

    @property
    def objective(self):
        self._append_cmd("objectives")
        return self._ScoreboardObjective(self)

    @property
    def players(self):
        self._append_cmd("players")
        return self._ScoreboardPlayers(self)


class Setblock(Command):
    def __init__(
        self,
        tile: str | Block,
        position: position = ("~", "~", "~"),
        **properties: str,
    ) -> None:
        super().__init__("setblock")

        name = (
            tile.identifier
            if isinstance(tile, (Blocks._MinecraftBlock, Block))
            else tile if isinstance(tile, str) else CONFIG.Logger.unsupported_block_type(tile)
        )
        states = (
            tile.states
            if isinstance(tile, Blocks._MinecraftBlock)
            else "" if isinstance(tile, (str, Block)) else CONFIG.Logger.unsupported_block_type(tile)
        )

        for k, v in properties.items():
            states.append(f'"{k}" = "{v}"')

        self._append_cmd(*position, name, f'[{", ".join(states)}]')


class Fill(Command):
    def __init__(
        self,
        tile: str | Block,
        start: position = ("~", "~", "~"),
        end: position = ("~", "~", "~"),
        old_block_handling: FillMode = FillMode.Replace,
        **properties: str,
    ) -> None:
        super().__init__("fill")

        name = (
            tile.identifier
            if isinstance(tile, (Blocks._MinecraftBlock, Block))
            else tile if isinstance(tile, str) else CONFIG.Logger.unsupported_block_type(tile)
        )
        states = (
            tile.states
            if isinstance(tile, Blocks._MinecraftBlock)
            else "" if isinstance(tile, (str, Block)) else CONFIG.Logger.unsupported_block_type(tile)
        )

        for k, v in properties.items():
            states.append(f'"{k}" = "{v}"')

        self._append_cmd(
            *start, *end, name, f'[{", ".join(states)}]', old_block_handling if old_block_handling != FillMode.Replace else ""
        )

    def replace(
        self,
        tile: str | Block,
        **properties: str,
    ):
        name = (
            tile.identifier
            if isinstance(tile, (Blocks._MinecraftBlock, Block))
            else tile if isinstance(tile, str) else CONFIG.Logger.unsupported_block_type(tile)
        )
        states = (
            tile.states
            if isinstance(tile, Blocks._MinecraftBlock)
            else "" if isinstance(tile, (str, Block)) else CONFIG.Logger.unsupported_block_type(tile)
        )

        for k, v in properties.items():
            states.append(f'"{k}" = "{v}"')

        self._append_cmd(FillMode.Replace, name, f'[{", ".join(states)}]')


class Music(Command):
    def __init__(self) -> None:
        super().__init__("music")

    def _base(
        self, track_name: str, volume: float = 1, fade_seconds: float = 1, repeat_mode: MusicRepeatMode = MusicRepeatMode.Once
    ):
        self._append_cmd(track_name)
        if volume != 1:
            self._append_cmd(volume)

        if fade_seconds != 1:
            self._append_cmd(fade_seconds)

        if repeat_mode != MusicRepeatMode.Once:
            self._append_cmd(repeat_mode)

        return self

    def queue(
        self, track_name: str, volume: float = 1, fade_seconds: float = 1, repeat_mode: MusicRepeatMode = MusicRepeatMode.Once
    ):
        self._append_cmd("queue")
        return self._base(track_name, volume, fade_seconds, repeat_mode)

    def play(
        self, track_name: str, volume: float = 1, fade_seconds: float = 1, repeat_mode: MusicRepeatMode = MusicRepeatMode.Once
    ):
        self._append_cmd("play")
        return self._base(track_name, volume, fade_seconds, repeat_mode)

    def stop(self, fade_seconds: float = 1):
        self._append_cmd("stop")

        if fade_seconds != 1:
            self._append_cmd(fade_seconds)

        return self

    def volume(self, volume: float):
        self._append_cmd("volume", volume)

        return self


class ScriptEvent(Command):
    def __init__(self, message_id: str, message: str = None) -> None:
        """Causes an event to fire within script with the specified message ID and payload. This can be used to connect scripting with any location where commands are used.

        Args:
            message_id (str): Identifier of the message to send. This is custom and dependent on the kinds of behavior packs and content you may have installed within the world.
            message (str): Data component of the message to send. This is custom and dependent on the kinds of behavior packs and content you may have installed within the world.

        [Documentation reference]: https://learn.microsoft.com/en-gb/minecraft/creator/commands/commands/scriptevent
        """
        super().__init__("scriptevent", message_id, message if message != None else "")


class Camera(Command):
    class _CameraSet(Command):
        def __init__(self, target, preset) -> None:
            super().__init__("camera", target, "set", preset)

            self.argument = {
                "easing": None,
                "easing_time": None,
                "position": None,
                "rotation": None,
                "facing_target": None,
                "facing_position": None,
            }

        def ease(self, easing: CameraEasing, easing_time: float):
            self.argument["easing_time"] = easing_time
            self.argument["easing"] = easing
            return self

        def position(self, position: coordinates):
            self.argument["position"] = position
            return self

        def rotation(self, xrotation: float, yrotation: float):
            self.argument["rotation"] = (xrotation, yrotation)
            return self

        def facing_target(self, target: Selector | Target):
            self.argument["facing_target"] = target
            return self

        def facing_position(self, position: coordinates):
            self.argument["facing_position"] = position
            return self

        def __str__(self):
            rotation = self.argument["rotation"]
            facing_target = self.argument["facing_target"]
            facing_position = self.argument["facing_position"]

            if sum([x is not None for x in (facing_target, facing_position, rotation)]) > 1:
                CONFIG.Logger.multiple_rotations()

            if self.argument["easing"] != None:
                self._append_cmd("ease", self.argument["easing_time"], self.argument["easing"])

            if self.argument["position"] != None:
                self._append_cmd("pos", *self.argument["position"])

            if self.argument["rotation"] != None:
                self._append_cmd("rot", *self.argument["rotation"])

            if self.argument["facing_target"] != None:
                self._append_cmd("facing", self.argument["facing_target"])

            if self.argument["facing_position"] != None:
                self._append_cmd("facing", *self.argument["facing_position"])

            return super().__str__()

    class _CameraClear(Command):
        def __init__(self, target) -> None:
            super().__init__("camera", target, "clear")

    class _CameraFade(Command):
        def __init__(self, target) -> None:
            super().__init__("camera", target, "fade")

            self.argument = {
                "fade_in_time": None,
                "hold_time": None,
                "fade_out_time": None,
                "color": None,
            }

        def time(self, fade_in_time: Seconds, hold_time: Seconds, fade_out_time: Seconds):
            self.argument["fade_in_time"] = fade_in_time
            self.argument["hold_time"] = hold_time
            self.argument["fade_out_time"] = fade_out_time
            return self

        def color(self, color: Color):
            self.argument["color"] = color
            return self

        def __str__(self):
            if self.argument["fade_in_time"] != None:
                self._append_cmd(self.argument["fade_in_time"], self.argument["hold_time"], self.argument["fade_out_time"])

            if self.argument["color"] != None:
                self._append_cmd(self.argument["color"])

            return super().__str__()

    def __init__(self, target: Target | Selector) -> None:
        """Transforms the camera for the selected player to a different perspective."""
        self._target = target

    def set(self, preset: CameraPresets):
        return self._CameraSet(self._target, preset)

    @property
    def clear(self):
        return self._CameraClear(self._target)

    @property
    def fade(self):
        return self._CameraFade(self._target)


class Time(Command):
    def __init__(self) -> None:
        """Changes the world's game time.

        [Documentation reference]: https://learn.microsoft.com/en-gb/minecraft/creator/commands/commands/time
        """
        super().__init__("time")

    def add(self, amount: int):
        """Add an integer amount of time in-game.

        Args:
            amount (int): The amount of time to add.
        """
        self._append_cmd("add", amount)
        return self

    def set(self, amount: TimeSpec | int):
        """Set the time in-game.

        Args:
            amount (TimeSpec | int): The time to set.
        """
        self._append_cmd("set", amount)
        return self


class Stopsound(Command):
    def __init__(self, target: Target, sound: str = "") -> None:
        """Stop a sound for a player with an optional sound string .

        Args:
            target (Target): A player name string or Target to identify the player.
            sound (str, optional): A string from the sound enum of the sound to stop. Defaults to "".

        [Documentation reference]: https://learn.microsoft.com/en-gb/minecraft/creator/commands/commands/stopsound
        """
        super().__init__("stopsound", target)

        if not sound == "":
            self._append_cmd(f'"{sound}"')


class Hud(Command):
    def __init__(
        self, target: Selector | Target, visible: HudVisibility = HudVisibility.Reset, hud_element: HudElement = HudElement.All
    ):
        """Changes the visibility of HUD elements for a player.

        Args:
            target (Selector | Target): The player to change the HUD visibility for.
            visible (HudVisibility, optional): The visibility state to set. Defaults to HudVisibility.Reset.
            hud_element (HudElement, optional): The HUD element to change the visibility of. Defaults to HudElement.All.

        [Documentation reference]: https://learn.microsoft.com/en-gb/minecraft/creator/commands/commands/hud
        """

        super().__init__("hud", target, visible, hud_element)
        return self



import os
from enum import StrEnum

from anvil import ANVIL, CONFIG
from anvil.api.actors import _Components
from anvil.api.blockbench import _Blockbench
from anvil.api.components import _component
from anvil.api.enums import (BlockFaces, BlockMaterial, BlockVanillaTags,
                             ItemCategory, ItemGroups, PlacementDirectionTrait,
                             PlacementPositionTrait)
from anvil.api.types import Molang, coordinates, position
from anvil.lib.format_versions import BLOCK_SERVER_VERSION, ITEM_SERVER_VERSION
from anvil.lib.lib import CopyFiles, FileExists, clamp
from anvil.lib.reports import ReportType
from anvil.lib.schemas import AddonObject, JsonSchemes, MinecraftDescription

__all__ = [
    "Block",
    "VanillaBlockTexture",
    "BlockDestructibleByExplosion",
    "BlockDestructibleByMining",
    "BlockFlammable",
    "BlockFriction",
    "BlockLightDampening",
    "BlockLightEmission",
    "BlockLootTable",
    "BlockMapColor",
    "BlockMaterialInstance",
    "BlockGeometry",
    "BlockCollisionBox",
    "BlockSelectionBox",
    "BlockPlacementFilter",
    "BlockTransformation",
    "BlockDisplayName",
    "BlockCraftingTable",
    "PlacementDirectionTrait",
    "PlacementPositionTrait",
    "CardinalDirectionsTrait",
    "FacingDirectionsTrait",
    "BlockFacesTrait",
    "VerticalHalfTrait",
]


class BlockDescriptor(dict):
    """
    A class that inherits from Python's built-in dict class. It is used to create a descriptor for a block in Minecraft
    with its name, tags, and states.
    """

    def __init__(self, name: str, tags: Molang, **states):
        """
        Initializes a BlockDescriptor object.

        Args:
            name (str): The name of the block.
            tags (Molang): The tags of the block.
            **states: The states of the block.
        """
        super().__init__(name=name, tags=tags, states=states)


# Components
class BlockRedstoneConductivity(_component):
    component_namespace = "minecraft:redstone_conductivity"

    def __init__(self, allows_wire_to_step_down: bool = True, redstone_conductor: bool = False) -> None:
        """Specifies whether a block has redstone properties. If the component is not provided, the default values are used.

        Args:
            allows_wire_to_step_down (bool, optional): Specifies if redstone wire can stair-step downward on the block, Defaults to True.
            redstone_conductor (bool, optional): Specifies if the block can be powered by redstone, Defaults to False.

        [Documentation reference]: https://learn.microsoft.com/en-gb/minecraft/creator/reference/content/blockreference/examples/blockcomponents/minecraft_redstone_conductivity
        """
        super().__init__("redstone_conductivity")
        self._enforce_version(BLOCK_SERVER_VERSION, "1.21.30")
        self._component_add_field("allows_wire_to_step_down", allows_wire_to_step_down)
        self._component_add_field("redstone_conductor", redstone_conductor)


class BlockCustomComponents(_component):
    component_namespace = "minecraft:custom_components"

    def __init__(self, *components: str) -> None:
        """Sets the color of the block when rendered to a map.

        Args:
            components (str): The components to register, if the namespace is not provided, the project namespace will be used.

        [Documentation reference]: https://learn.microsoft.com/en-gb/minecraft/creator/reference/content/blockreference/examples/blockcomponents/minecraft_custom_components
        """
        super().__init__("custom_components")
        self._component_set_value(components)


class BlockDefault(_component):
    component_namespace = "minecraft:block_default"

    def __init__(self) -> None:
        """The default block component."""
        super().__init__("block_default")

    def brightness_gamma(self, gamma: int):
        """The gamma of the block's brightness."""
        self._component_add_field("brightness_gamma", max(0, gamma))
        return self

    def sound(self, sound: str):
        """The sound of the block."""
        self._component_add_field("sound", sound)
        return self

    def isotropic(self, **kwargs: dict[str, BlockFaces]):
        """The isotropic of the block."""
        if BlockFaces.All in kwargs.values():
            self._component_set_value("isotropic", True)
        else:
            self._component_add_field("isotropic", {v: k for k, v in kwargs.items()})
        return self

    def textures(self, **kwargs: dict[str, BlockFaces]):
        """The textures of the block."""
        for k in kwargs.keys():
            if not FileExists(os.path.join("assets", "textures", "blocks", f"{k}.png")):
                CONFIG.Logger.file_exist_error(f"{k}.png", os.path.join("assets", "textures", "blocks"))

            CopyFiles(
                os.path.join("assets", "textures", "blocks"),
                os.path.join(CONFIG.RP_PATH, "textures", CONFIG.NAMESPACE, CONFIG.PROJECT_NAME, "blocks").replace("\\", "/"),
                f"{k}.png",
            )
            ANVIL.definitions.register_terrain_texture(k, "", k)

        if BlockFaces.All in kwargs.values():
            self._component_add_field(
                "textures", [f"{CONFIG.NAMESPACE}:{k}" for k, v in kwargs.items() if v == BlockFaces.All][0]
            )
        else:
            self._component_add_field("textures", {v: f"{CONFIG.NAMESPACE}:{k}" for k, v in kwargs.items()})

        return self

    def carried_textures(self, **kwargs: dict[str, BlockFaces]):
        """The carried textures of the block."""
        if BlockFaces.All in kwargs.values():
            self._component_set_value(
                "carried_textures", [f"{CONFIG.NAMESPACE}:{k}" for k, v in kwargs.items() if v == BlockFaces.All][0]
            )
        else:
            self._component_add_field("carried_textures", {v: f"{CONFIG.NAMESPACE}:{k}" for k, v in kwargs.items()})
        return self


class BlockDestructibleByExplosion(_component):
    component_namespace = "minecraft:destructible_by_explosion"

    def __init__(self, explosion_resistance: int) -> None:
        """Describes the destructible by explosion properties for this block.

        Args:
            explosion_resistance (int): The amount of resistance to explosions in a range of 0 to 100.
        """
        super().__init__("destructible_by_explosion")
        self._component_add_field("explosion_resistance", explosion_resistance)


class BlockDestructibleByMining(_component):
    component_namespace = "minecraft:destructible_by_mining"

    def __init__(self, seconds_to_destroy: int) -> None:
        """Describes the destructible by mining properties for this block.

        Args:
            seconds_to_destroy (int): The amount of time it takes to destroy the block in seconds.
        """
        super().__init__("destructible_by_mining")
        self._enforce_version(ITEM_SERVER_VERSION, "1.21.50")
        self._component_add_field("seconds_to_destroy", seconds_to_destroy)
        self._component_add_field("item_specific_speeds", [])

    def item_specific_speeds_name(self, destroy_speed: float, item_name: str):
        self[self.component_namespace]["item_specific_speeds"].append({"item": item_name, "destroy_speed": destroy_speed})

    def item_specific_speeds_tag(self, destroy_speed: float, item_tag: str | Molang):
        self[self.component_namespace]["item_specific_speeds"].append(
            {"item": {"tags": item_tag}, "destroy_speed": destroy_speed}
        )


class BlockFlammable(_component):
    component_namespace = "minecraft:flammable"

    def __init__(self, catch_chance_modifier: int, destroy_chance_modifier: int) -> None:
        """Describes the flammable properties for this block.

        Args:
            catch_chance_modifier (int): The chance that this block will catch fire in a range of 0 to 100.
            destroy_chance_modifier (int): The chance that this block will be destroyed when on fire in a range of 0 to 100.
        """
        super().__init__("flammable")
        self._component_add_field("catch_chance_modifier", catch_chance_modifier)
        self._component_add_field("destroy_chance_modifier", destroy_chance_modifier)


class BlockFriction(_component):
    component_namespace = "minecraft:friction"

    def __init__(self, friction: float = 0.4) -> None:
        """Describes the friction for this block in a range of 0.0 to 0.9.

        Args:
            friction (float, optional): The friction of the block. Defaults to 0.4.
        """
        super().__init__("flammable")
        self._component_set_value(clamp(friction, 0, 0.9))


class BlockLightDampening(_component):
    component_namespace = "minecraft:light_dampening"

    def __init__(self, light_dampening: int = 15) -> None:
        """The amount that light will be dampened when it passes through the block in a range of 0 to 15.

        Args:
            light_dampening (int, optional): The amount of light dampening. Defaults to 15.
        """
        super().__init__("light_dampening")
        self._component_set_value(clamp(light_dampening, 0, 15))


class BlockLightEmission(_component):
    component_namespace = "minecraft:light_emission"

    def __init__(self, light_emission: int = 0) -> None:
        """The amount of light this block will emit in a range of 0 to 15.

        Args:
            light_emission (int, optional): The amount of light emission. Defaults to 0.
        """
        super().__init__("light_emission")
        self._component_set_value(clamp(light_emission, 0, 15))


class BlockLootTable(_component):
    component_namespace = "minecraft:loot"

    def __init__(self, loot: str) -> None:
        """Specifies the path to the loot table.

        Args:
            loot (str): The path to the loot table.
        """
        super().__init__("loot")
        self._component_set_value(loot)


class BlockMapColor(_component):
    component_namespace = "minecraft:map_color"

    def __init__(self, map_color: str) -> None:
        """Sets the color of the block when rendered to a map.

        Args:
            map_color (str): The color of the block when rendered to a map, must be a hexadecimal color value.
        """
        super().__init__("map_color")
        self._component_set_value(map_color)


class BlockMaterialInstance(_component):
    component_namespace = "minecraft:material_instances"

    def __init__(self) -> None:
        """Maps face or material_instance names in a geometry file to an actual material instance."""
        super().__init__("material_instances")

    def add_instance(
        self,
        blockbench_name: str,
        texture_name: str,
        block_face: BlockFaces = BlockFaces.All,
        render_method: BlockMaterial = BlockMaterial.Opaque,
        ambient_occlusion: bool = True,
        face_dimming: bool = True,
    ):
        """Maps face or material_instance names in a geometry file to an actual material instance.

        Args:
            texture_name (str): The name of the texture to use for this block.
            block_face (BlockFaces, optional): The face of the block to apply the texture to. Defaults to BlockFaces.All.
            render_method (BlockMaterial, optional): The render method to use for this block. Defaults to BlockMaterial.Opaque.
            ambient_occlusion (bool, optional): Whether or not to use ambient occlusion for this block. Defaults to True.
            face_dimming (bool, optional): Whether or not to use face dimming for this block. Defaults to True.

        """
        bb = _Blockbench(blockbench_name, "blocks")
        bb.textures.queue_texture(texture_name)

        ANVIL.definitions.register_terrain_texture(texture_name, blockbench_name, texture_name)

        self[self.component_namespace].update(
            {
                "*" if block_face == BlockFaces.All else block_face: {
                    "texture": f"{CONFIG.NAMESPACE}:{texture_name}",
                    "render_method": render_method if not render_method == BlockMaterial.Opaque else {},
                    "ambient_occlusion": ambient_occlusion if ambient_occlusion is False else {},
                    "face_dimming": face_dimming if face_dimming is False else {},
                }
            }
        )

        return self


class BlockGeometry(_component):
    component_namespace = "minecraft:geometry"

    def __init__(self, geometry_name: str) -> None:
        """The description identifier of the geometry file to use to render this block.

        Args:
            geometry_name (str): The name of the geometry to use to render this block.
        """
        super().__init__("geometry")
        self._component_set_value(f"geometry.{CONFIG.NAMESPACE}.{geometry_name}")

        bb = _Blockbench(geometry_name, "blocks")
        bb.model.queue_model()

    def bone_visibility(self, **bone: dict[str, bool | str | Molang]):
        """Specifies the visibility of bones in the geometry file.

        Example:
            >>> BlockGeometry('block').bone_visibility(bone0=True, bone1=False)

        """
        self._component_add_field("bone_visibility", [{b: v for b, v in bone.items()}])
        return self


class BlockCollisionBox(_component):
    component_namespace = "minecraft:collision_box"

    def __init__(self, size: coordinates, origin: coordinates) -> None:
        """Defines the area of the block that collides with entities.

        Args:
            size (coordinates): The size of the collision box.
            origin (coordinates): The origin of the collision box.
        """
        super().__init__("collision_box")
        if size == (0, 0, 0):
            self._component_set_value(False)
        else:
            self._component_add_field("size", size)
            self._component_add_field("origin", origin)


class BlockSelectionBox(_component):
    component_namespace = "minecraft:selection_box"

    def __init__(self, size: coordinates, origin: coordinates) -> None:
        """Defines the area of the block that is selected by the player's cursor.

        Args:
            size (coordinates): The size of the selection box.
            origin (coordinates): The origin of the selection box.
        """
        super().__init__("selection_box")
        if size == (0, 0, 0):
            self._component_set_value(False)
        else:
            self._component_add_field("size", size)
            self._component_add_field("origin", origin)


class BlockPlacementFilter(_component):
    component_namespace = "minecraft:placement_filter"

    def __init__(self) -> None:
        """By default, custom blocks can be placed anywhere and do not have placement restrictions unless you specify them in this component."""
        super().__init__("placement_filter")
        self._component_add_field("conditions", [])

    def add_condition(self, allowed_faces: list[BlockFaces], block_filter: list[BlockDescriptor | str]):
        """Adds a condition to the placement filter.

        Args:
            allowed_faces (list[BlockFaces]): The faces of the block that are allowed to be placed on.
            block_filter (list[BlockDescriptor | str]): The blocks that are allowed to be placed on.
        """
        if BlockFaces.Side in allowed_faces:
            allowed_faces.remove(BlockFaces.North)
            allowed_faces.remove(BlockFaces.South)
            allowed_faces.remove(BlockFaces.East)
            allowed_faces.remove(BlockFaces.West)
        if BlockFaces.All in allowed_faces:
            allowed_faces = [BlockFaces.All]
        self[self.component_namespace]["conditions"].append(
            {
                "allowed_faces": allowed_faces,
                "block_filter": block_filter,
            }
        )
        return self


class BlockTransformation(_component):
    component_namespace = "minecraft:transformation"

    def __init__(self) -> None:
        """The block's transfomration."""
        super().__init__("transformation")

    def translation(self, translation: tuple[float, float, float]):
        """The block's translation.

        Args:
            translation (position): The block's translation.
        """
        self._component_add_field("translation", translation)
        return self

    def scale(self, scale: tuple[float, float, float], scale_pivot: tuple[float, float, float] = (0, 0, 0)):
        """The block's scale.

        Args:
            scale (position): The block's scale.
        """
        self._component_add_field("scale", scale)

        if scale_pivot != (0, 0, 0):
            self._component_add_field("scale_pivot", scale_pivot)

        return self

    def rotation(self, rotation: tuple[float, float, float], rotation_pivot: tuple[float, float, float] = (0, 0, 0)):
        """The block's rotation.

        Args:
            rotation (position): The block's rotation.

        """
        self._component_add_field("rotation", rotation)

        if rotation_pivot != (0, 0, 0):
            self._component_add_field("rotation_pivot", rotation_pivot)
        return self


class BlockDisplayName(_component):
    component_namespace = "minecraft:display_name"

    def __init__(self, display_name: str, localize: bool = True) -> None:
        """Sets the block display name within Minecraft: Bedrock Edition. This component may also be used to pull from the localization file by referencing a key from it.

        Args:
            display_name (str): Set the display name for an block.
            localize (bool, optional): Whether to use the name with a localization file or not. Defaults to True.

        """
        super().__init__("display_name")
        if localize:
            key = f'tile.{CONFIG.NAMESPACE}:{display_name.lower().replace(" ", "_")}.name'
            ANVIL.definitions.register_lang(key, display_name)
            self._component_set_value(key)
        else:
            self._component_set_value(display_name)


class BlockCraftingTable(_component):
    component_namespace = "minecraft:crafting_table"

    def __init__(self, table_name: str, *crafting_tags: str) -> None:
        """Makes your block into a custom crafting table which enables the crafting table UI and the ability to craft recipes.

        Args:
            table_name (str): The name of the crafting table.

        Raises:
            IndexError: The crafting table tags cannot exceed 64 tags.
            ValueError: The crafting table tags are limited to 64 characters.
        """
        super().__init__("crafting_table")
        self._component_add_field("table_name", table_name)

        if len(crafting_tags) > 64:
            raise IndexError("Crafting Table tags cannot exceed 64 tags.")

        for tag in crafting_tags:
            if len(tag) > 64:
                raise ValueError("Crafting Table tags are limited to 64 characters.")

        self._component_add_field("crafting_tags", crafting_tags)


# Core
class _PermutationComponents(_Components):
    _count = 0

    def __init__(self, condition: str | Molang = None):
        """The permutation components.

        Args:
            condition (str | Molang): The condition for the permutation.
        """
        super().__init__()
        _PermutationComponents._count += 1
        self._component_group_name = "components"
        self._condition = condition

    def tag(self, *tags: BlockVanillaTags):
        """The tags for the block.

        Args:
            tags (BlockVanillaTags): The tags for the block.

        [Documentation reference]: https://learn.microsoft.com/en-gb/minecraft/creator/reference/content/blockreference/examples/blocktags
        """
        for tag in tags:
            self._set(f"tag:{tag}", {f"tag:{tag}": {"do_not_shorten": True}})

        return self

    def _export(self):
        cmp = super()._export()
        if self._condition:
            cmp["condition"] = self._condition
        return cmp


class _BlockTraits:
    def __init__(self) -> None:
        self._traits = {}

    def placement_direction(self, y_rotation_offset: float = 0, *traits: PlacementDirectionTrait):
        """can add states containing information about the player's rotation when the block is placed.

        Args:
            y_rotation_offset (float, optional): The y rotation offset. Defaults to 0.
            traits (PlacementDirectionTrait): The traits for the block.

        [Documentation reference]: https://learn.microsoft.com/en-gb/minecraft/creator/reference/content/blockreference/examples/blocktraits#placement_direction-example
        """

        self._traits["minecraft:placement_direction"] = {
            "enabled_states": [t for t in traits if t],
            "y_rotation_offset": y_rotation_offset,
        }

    def placement_position(self, *traits: PlacementPositionTrait):
        """Can add states containing information about the position of the block when it is placed.

        Args:
            traits (PlacementPositionTrait): The traits for the block.

        [Documentation reference]: https://learn.microsoft.com/en-gb/minecraft/creator/reference/content/blockreference/examples/blocktraits#placement_position-example
        """
        self._traits["minecraft:placement_position"] = {"enabled_states": traits}

    @property
    def export(self):
        return self._traits


class _BlockServerDescription(MinecraftDescription):
    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """The block description.

        Args:
            name (str): The name of the block.
            is_vanilla (bool, optional): Whether or not the block is a vanilla block. Defaults to False.
        """
        super().__init__(name, is_vanilla)
        self._traits = _BlockTraits()
        self._description["description"].update(
            {
                "states": {},
                "traits": {},
            }
        )

    def add_state(self, name: str, *range: float | str | bool):
        """Adds a state to the block.

        Args:
            name (str): The name of the state.
            range (float | str |bool): Values this state can have.

        """
        if len(range) > 16:
            CONFIG.Logger.block_state_values_out_of_range(self._name, name, len(range))
        self._description["description"]["states"][f"{CONFIG.NAMESPACE}:{name}"] = range
        return self

    def menu_category(
        self,
        category: ItemCategory = ItemCategory.none,
        group: ItemGroups | str = ItemGroups.none,
    ):
        """Sets the menu category for the block.

        Args:
            category (ItemCategory, optional): The category of the block. Defaults to ItemCategory.none.
            group (str, optional): The group of the block. Defaults to None.

        """
        self._description["description"]["menu_category"] = {
            "category": category.value if not category == ItemCategory.none else {},
            "group": group if not group == ItemGroups.none else {},
        }
        return self

    @property
    def is_hidden_in_commands(self):
        self._description["description"]["is_hidden_in_commands"] = True
        return self

    @property
    def traits(self):
        """Sets the traits for the block.

        [Documentation reference]: https://learn.microsoft.com/en-gb/minecraft/creator/reference/content/blockreference/examples/blocktraits
        """
        return self._traits

    @property
    def to_dict(self):
        self._description["description"]["traits"] = self._traits.export
        return super().to_dict


class _BlockServer(AddonObject):
    """The block server object."""

    _extension = ".block.json"
    _path = os.path.join(CONFIG.BP_PATH, "blocks")

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """The block server object.

        Args:
            name (str): The name of the block.
            is_vanilla (bool, optional): Whether or not the block is a vanilla block. Defaults to False.
        """
        super().__init__(name)
        self._server_block = JsonSchemes.server_block()
        self._description = _BlockServerDescription(name, is_vanilla)
        self._components = _PermutationComponents(None)
        self._permutations: list[_PermutationComponents] = []

    @property
    def description(self):
        """The block description."""
        return self._description

    @property
    def components(self):
        """The block components."""
        return self._components

    def permutation(self, condition: str | Molang):
        """Adds a permutation to the block.

        Args:
            condition (str | Molang): The condition for the permutation.
        """
        self._permutation = _PermutationComponents(condition)
        self._permutations.append(self._permutation)
        return self._permutation

    @property
    def queue(self):
        """Queues the block to be exported."""
        self._server_block["minecraft:block"].update(self.description.to_dict)
        self._server_block["minecraft:block"].update(self._components._export())
        comps: dict = self._server_block["minecraft:block"]["components"]
        self._server_block["minecraft:block"]["permutations"] = [permutation._export() for permutation in self._permutations]

        if not BlockDefault.component_namespace in comps:
            if not BlockMaterialInstance.component_namespace in comps:
                CONFIG.Logger.block_missing_texture(self._name)
            if not BlockGeometry.component_namespace in comps:
                CONFIG.Logger.block_missing_geometry(self._name)
        else:
            ANVIL.definitions.register_block(self.description.identifier, comps[BlockDefault.component_namespace])
            comps.pop(BlockDefault.component_namespace)

        if not BlockDisplayName.component_namespace in self._server_block["minecraft:block"]["components"]:
            display_name = self._name.replace("_", " ").title()
            self._server_block["minecraft:block"]["components"][BlockDisplayName.component_namespace] = display_name

        self.content(self._server_block)
        super().queue()


class Block:
    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """The block object.

        Args:
            name (str): The name of the block.
            is_vanilla (bool, optional): Whether or not the block is a vanilla block. Defaults to False.
        """
        self._name = name
        self._is_vanilla = is_vanilla
        self._server = _BlockServer(name, is_vanilla)

        self._namespace_format = CONFIG.NAMESPACE
        if self._is_vanilla:
            self._namespace_format = "minecraft"

    @property
    def Server(self):
        """The block server object."""
        return self._server

    @property
    def identifier(self):
        """The block identifier."""
        return f"{self._namespace_format}:{self._name}"

    @property
    def name(self):
        return self._name

    def queue(self):
        """Queues the block to be exported."""
        self.Server.queue

        if self.Server._server_block["minecraft:block"]["components"][BlockDisplayName.component_namespace].startswith("tile."):
            display_name = ANVIL.definitions._language[
                self.Server._server_block["minecraft:block"]["components"][BlockDisplayName.component_namespace]
            ]
        else:
            display_name = self.Server._server_block["minecraft:block"]["components"][BlockDisplayName.component_namespace]

        CONFIG.Report.add_report(
            ReportType.BLOCK,
            vanilla=self._is_vanilla,
            col0=display_name,
            col1=self.identifier,
            col2=[
                f"{key}: {[', '.join(str(v) for v in value)]}"
                for key, value in self.Server.description._description["description"]["states"].items()
            ],
        )


#  TODO: Integrate with the Block class


class VanillaBlockTexture(MinecraftDescription):
    def __init__(self, identifier: str) -> None:
        super().__init__(identifier, True)
        self._identifier = identifier

    @property
    def update_texture(self):
        self._ext = ""
        if FileExists(os.path.join("assets", "textures", "blocks", f"{self._identifier}.png")):
            self._ext = "png"
        if FileExists(os.path.join("assets", "textures", "blocks", f"{self._identifier}.tga")):
            self._ext = "tga"

        return self

    def queue(self, directory: str = ""):
        CopyFiles(
            os.path.join("assets", "textures", "blocks"),
            os.path.join(CONFIG.RP_PATH, "textures", "blocks", directory),
            f"{self._identifier}.{self._ext}",
        )
